"""Test the flow accumulator component.

@author: krb
"""
# Created on Thurs Nov 12, 2015
import os

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import HexModelGrid, RasterModelGrid
from landlab.components import LinearDiffuser
from landlab.components.flow_accum import LossyFlowAccumulator
from landlab.components.flow_director import (
    FlowDirectorD8,
    FlowDirectorDINF,
    FlowDirectorMFD,
    FlowDirectorSteepest,
)
from landlab.components.flow_routing.lake_mapper import DepressionFinderAndRouter

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_loss_func_arguments():
    """Check the loss_function only has one argument."""
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x ** 2 + mg.node_y ** 2, at="node")

    def funcgood(x):
        return 0.0

    def funcgood2(x, y):
        return 0.0

    def funcgood3(x, y, z):
        return 0.0

    def funcgood4(x, y, z, grid):
        return 0.0

    def funcbad(x, y, z, grid, K):
        return 0.0

    def funcbad2(x):
        return "booooooo"

    def funcbad3(x, y):
        return np.array([0.0])

    def funcbad4(x, y, z):
        return (0, 1)

    LossyFlowAccumulator(mg, loss_function=funcgood)  # no problem
    LossyFlowAccumulator(mg, loss_function=funcgood2)  # no problem
    LossyFlowAccumulator(mg, loss_function=funcgood3)  # no problem
    LossyFlowAccumulator(mg, loss_function=funcgood4)  # no problem
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, loss_function=funcbad)
    with pytest.raises(TypeError):
        LossyFlowAccumulator(mg, loss_function=funcbad2)
    with pytest.raises(TypeError):
        LossyFlowAccumulator(mg, loss_function=funcbad3)
    with pytest.raises(TypeError):
        LossyFlowAccumulator(mg, loss_function=funcbad4)


def test_run_with_2_fn_args():
    mg = RasterModelGrid((3, 5), xy_spacing=(2, 1))
    mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")

    def mylossfunction(qw, nodeID):
        return 0.5 * qw

    fa = LossyFlowAccumulator(
        mg,
        "topographic__elevation",
        flow_director=FlowDirectorSteepest,
        routing="D4",
        loss_function=mylossfunction,
    )
    fa.run_one_step()

    assert np.allclose(
        mg.at_node["drainage_area"].reshape(mg.shape),
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [6.0, 6.0, 4.0, 2.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        ),
    )
    assert np.allclose(
        mg.at_node["surface_water__discharge"].reshape(mg.shape),
        np.array(
            [
                [0.00, 0.00, 0.00, 0.00, 0.00],
                [1.75, 3.50, 3.00, 2.00, 0.00],
                [0.00, 0.00, 0.00, 0.00, 0.00],
            ]
        ),
    )


def test_run_with_3_fn_args():
    mg = RasterModelGrid((3, 5), xy_spacing=(2, 1))
    mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")

    def mylossfunction(qw, nodeID, linkID):
        return 0.5 * qw

    fa = LossyFlowAccumulator(
        mg,
        "topographic__elevation",
        flow_director=FlowDirectorSteepest,
        routing="D4",
        loss_function=mylossfunction,
    )
    fa.run_one_step()

    assert np.allclose(
        mg.at_node["drainage_area"].reshape(mg.shape),
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [6.0, 6.0, 4.0, 2.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        ),
    )
    assert np.allclose(
        mg.at_node["surface_water__discharge"].reshape(mg.shape),
        np.array(
            [
                [0.00, 0.00, 0.00, 0.00, 0.00],
                [1.75, 3.50, 3.00, 2.00, 0.00],
                [0.00, 0.00, 0.00, 0.00, 0.00],
            ]
        ),
    )


def test_check_fields():
    """Check to make sure the right fields have been created."""

    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    z = mg.add_field(
        "topographic__elevation", mg.node_x ** 2 + mg.node_y ** 2, at="node"
    )

    LossyFlowAccumulator(mg)
    assert_array_equal(z, mg.at_node["topographic__elevation"])
    assert_array_equal(np.zeros(100), mg.at_node["drainage_area"])
    assert_array_equal(np.ones(100), mg.at_node["water__unit_flux_in"])
    assert_array_equal(np.zeros(100), mg.at_node["surface_water__discharge_loss"])

    LossyFlowAccumulator(mg, runoff_rate=2.0)
    assert_array_equal(np.full(100, 2.0), mg.at_node["water__unit_flux_in"])

    # quick test that the component binds correctly to an existing field:
    L = mg.at_node["surface_water__discharge_loss"]
    fa = LossyFlowAccumulator(mg)
    fa.run_one_step()  # this line is padding to make flake8 happy
    L[0] = 1.0
    assert mg.at_node["surface_water__discharge_loss"] is L


def test_director_adding_methods_are_equivalent_Steepest():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg0.add_field(
        "topographic__elevation", mg0.node_x ** 2 + mg0.node_y ** 2, at="node"
    )
    fa0 = LossyFlowAccumulator(mg0, flow_director="D4")
    fa0.run_one_step()

    mg1 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg1.add_field(
        "topographic__elevation", mg1.node_x ** 2 + mg1.node_y ** 2, at="node"
    )
    fa1 = LossyFlowAccumulator(mg1, flow_director="Steepest")
    fa1.run_one_step()

    mg2 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg2.add_field(
        "topographic__elevation", mg2.node_x ** 2 + mg2.node_y ** 2, at="node"
    )
    fa2 = LossyFlowAccumulator(mg2, flow_director=FlowDirectorSteepest)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg3.add_field(
        "topographic__elevation", mg3.node_x ** 2 + mg3.node_y ** 2, at="node"
    )
    fd = FlowDirectorSteepest(mg3)
    fa3 = LossyFlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key], mg1.at_node[key])

        assert_array_equal(mg1.at_node[key], mg2.at_node[key])

        assert_array_equal(mg2.at_node[key], mg3.at_node[key])


def test_director_adding_methods_are_equivalent_D8():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg0.add_field(
        "topographic__elevation", mg0.node_x ** 2 + mg0.node_y ** 2, at="node"
    )
    fa0 = LossyFlowAccumulator(mg0, flow_director="D8")
    fa0.run_one_step()

    mg1 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg1.add_field(
        "topographic__elevation", mg1.node_x ** 2 + mg1.node_y ** 2, at="node"
    )
    fa1 = LossyFlowAccumulator(mg1, flow_director="FlowDirectorD8")
    fa1.run_one_step()

    mg2 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg2.add_field(
        "topographic__elevation", mg2.node_x ** 2 + mg2.node_y ** 2, at="node"
    )
    fa2 = LossyFlowAccumulator(mg2, flow_director=FlowDirectorD8)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg3.add_field(
        "topographic__elevation", mg3.node_x ** 2 + mg3.node_y ** 2, at="node"
    )
    fd = FlowDirectorD8(mg3)
    fa3 = LossyFlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key], mg1.at_node[key])

        assert_array_equal(mg1.at_node[key], mg2.at_node[key])

        assert_array_equal(mg2.at_node[key], mg3.at_node[key])


def test_director_adding_methods_are_equivalent_Dinf():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg0.add_field(
        "topographic__elevation", mg0.node_x ** 2 + mg0.node_y ** 2, at="node"
    )
    fa0 = LossyFlowAccumulator(mg0, flow_director="DINF")
    fa0.run_one_step()

    mg1 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg1.add_field(
        "topographic__elevation", mg1.node_x ** 2 + mg1.node_y ** 2, at="node"
    )
    fa1 = LossyFlowAccumulator(mg1, flow_director="FlowDirectorDINF")
    fa1.run_one_step()

    mg2 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg2.add_field(
        "topographic__elevation", mg2.node_x ** 2 + mg2.node_y ** 2, at="node"
    )
    fa2 = LossyFlowAccumulator(mg2, flow_director=FlowDirectorDINF)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg3.add_field(
        "topographic__elevation", mg3.node_x ** 2 + mg3.node_y ** 2, at="node"
    )
    fd = FlowDirectorDINF(mg3)
    fa3 = LossyFlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key], mg1.at_node[key])

        assert_array_equal(mg1.at_node[key], mg2.at_node[key])

        assert_array_equal(mg2.at_node[key], mg3.at_node[key])


def test_director_adding_methods_are_equivalent_MFD():
    """Check that different methods to specifying the director are the same."""

    mg0 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg0.add_field(
        "topographic__elevation", mg0.node_x ** 2 + mg0.node_y ** 2, at="node"
    )
    fa0 = LossyFlowAccumulator(mg0, flow_director="MFD")
    fa0.run_one_step()

    mg1 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg1.add_field(
        "topographic__elevation", mg1.node_x ** 2 + mg1.node_y ** 2, at="node"
    )
    fa1 = LossyFlowAccumulator(mg1, flow_director="FlowDirectorMFD")
    fa1.run_one_step()

    mg2 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg2.add_field(
        "topographic__elevation", mg2.node_x ** 2 + mg2.node_y ** 2, at="node"
    )
    fa2 = LossyFlowAccumulator(mg2, flow_director=FlowDirectorMFD)
    fa2.run_one_step()

    mg3 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg3.add_field(
        "topographic__elevation", mg3.node_x ** 2 + mg3.node_y ** 2, at="node"
    )
    fd = FlowDirectorMFD(mg3)
    fa3 = LossyFlowAccumulator(mg3, flow_director=fd)
    fa3.run_one_step()

    for key in mg0.at_node.keys():
        assert_array_equal(mg0.at_node[key], mg1.at_node[key])

        assert_array_equal(mg1.at_node[key], mg2.at_node[key])

        assert_array_equal(mg2.at_node[key], mg3.at_node[key])


def test_passing_a_bad_component():
    """Check that a random component can't be a director."""
    from landlab.components import ChiFinder

    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, "topographic__elevation", flow_director=ChiFinder)


def test_error_for_to_many_with_depression():
    """Check that an error is thrown when to_many methods started DF."""

    mg0 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg0.add_field(
        "topographic__elevation", mg0.node_x ** 2 + mg0.node_y ** 2, at="node"
    )

    mg1 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg1.add_field(
        "topographic__elevation", mg1.node_x ** 2 + mg1.node_y ** 2, at="node"
    )

    with pytest.raises(NotImplementedError):
        LossyFlowAccumulator(
            mg0, flow_director="MFD", depression_finder="DepressionFinderAndRouter"
        )
    with pytest.raises(NotImplementedError):
        LossyFlowAccumulator(
            mg0, flow_director="DINF", depression_finder="DepressionFinderAndRouter"
        )

    fa0 = LossyFlowAccumulator(mg0, flow_director="MFD")
    fa0.run_one_step()
    with pytest.raises(NotImplementedError):
        DepressionFinderAndRouter(mg0)

    fa1 = LossyFlowAccumulator(mg1, flow_director="DINF")
    fa1.run_one_step()
    with pytest.raises(NotImplementedError):
        DepressionFinderAndRouter(mg1)


def test_fields():
    """Check to make sure the right fields have been created.

    Check that the sizes are also correct.
    """
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = LossyFlowAccumulator(mg)
    fa.run_one_step()

    assert sorted(list(mg.at_node.keys())) == [
        "drainage_area",
        "flow__data_structure_delta",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__sink_flag",
        "flow__upstream_node_order",
        "surface_water__discharge",
        "surface_water__discharge_loss",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]
    assert sorted(list(mg.at_grid.keys())) == ["flow__data_structure_D"]

    mg2 = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg2.add_field("topographic__elevation", mg2.node_x + mg2.node_y, at="node")
    fa2 = LossyFlowAccumulator(mg2, flow_director="MFD")
    fa2.run_one_step()
    assert sorted(list(mg2.at_node.keys())) == [
        "drainage_area",
        "flow__data_structure_delta",
        "flow__link_to_receiver_node",
        "flow__receiver_node",
        "flow__receiver_proportions",
        "flow__sink_flag",
        "flow__upstream_node_order",
        "surface_water__discharge",
        "surface_water__discharge_loss",
        "topographic__elevation",
        "topographic__steepest_slope",
        "water__unit_flux_in",
    ]

    assert sorted(list(mg2.at_grid.keys())) == ["flow__data_structure_D"]


def test_accumulated_area_closes():
    """Check that accumulated area is area of core nodes."""

    fds = ["Steepest", "D8", "MFD", "DINF"]

    for fd in fds:
        mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
        mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
        fa = LossyFlowAccumulator(mg)
        fa.run_one_step()

        drainage_area = mg.at_node["drainage_area"]
        drained_area = np.sum(drainage_area[mg.boundary_nodes])
        core_area = np.sum(mg.cell_area_at_node[mg.core_nodes])
        assert drained_area == core_area


# def test_passing_unnecessary_kwarg():
#     """Test that passing a bad kwarg raises a ValueError."""
#     mg = RasterModelGrid((10,10), xy_spacing=(1, 1))
#     z = mg.add_field('topographic__elevation', mg.node_x + mg.node_y, at = 'node')
#     with pytest.raises(ValueError):
#         LossyFlowAccumulator(mg, bad_kwarg='woo')


def test_specifying_routing_method_wrong():
    """Test specifying incorrect method for routing compatability with DepressionFinderAndRouter."""
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")

    with pytest.raises(ValueError):
        LossyFlowAccumulator(
            mg, flow_director="D4", depression_finder="DepressionFinderAndRouter"
        )

    df = DepressionFinderAndRouter(mg)
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director="D4", depression_finder=df)


def test_field_name_array_float_case1():
    """Topography as field, runoff rate as float"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )
    mg.add_field("node", "topographic__elevation", topographic__elevation)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, "topographic__elevation", runoff_rate=10.0)
    assert_array_equal(
        mg.at_node["water__unit_flux_in"], 10.0 * np.ones(mg.size("node"))
    )

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            10.0,
            50.0,
            0.0,
            0.0,
            10.0,
            50.0,
            0.0,
            0.0,
            10.0,
            40.0,
            0.0,
            0.0,
            10.0,
            20.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_field_name_array_float_case2():
    """Topography as field, runoff rate as field name"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    runoff_rate = [
        1.0,
        1.0,
        1.0,
        1.0,
        2.0,
        2.0,
        2.0,
        2.0,
        3.0,
        3.0,
        3.0,
        3.0,
        4.0,
        4.0,
        4.0,
        4.0,
        5.0,
        5.0,
        5.0,
        5.0,
    ]

    mg.add_field("node", "topographic__elevation", topographic__elevation)
    mg.add_field("node", "runoff_rate", runoff_rate)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, "topographic__elevation", runoff_rate="runoff_rate")

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            2.0,
            16.0,
            0.0,  # KRB double checked these numbers by hand 5/15/18 - OK
            0.0,
            2.0,
            16.0,
            0.0,
            0.0,
            3.0,
            14.0,
            0.0,
            0.0,
            4.0,
            8.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_field_name_array_float_case3():
    """Topography as field, runoff rate as array"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    runoff_rate = [
        1.0,
        1.0,
        1.0,
        1.0,
        2.0,
        2.0,
        2.0,
        2.0,
        3.0,
        3.0,
        3.0,
        3.0,
        4.0,
        4.0,
        4.0,
        4.0,
        5.0,
        5.0,
        5.0,
        5.0,
    ]

    mg.add_field("node", "topographic__elevation", topographic__elevation)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, "topographic__elevation", runoff_rate=runoff_rate)

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            2.0,
            16.0,
            0.0,  # KRB double checked these numbers by hand 5/15/18 - OK
            0.0,
            2.0,
            16.0,
            0.0,
            0.0,
            3.0,
            14.0,
            0.0,
            0.0,
            4.0,
            8.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_field_name_array_float_case4():
    """Topography as array, runoff rate as float"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )
    mg.add_field("node", "topographic__elevation", topographic__elevation)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, topographic__elevation, runoff_rate=10.0)
    assert_array_equal(
        mg.at_node["water__unit_flux_in"], 10.0 * np.ones(mg.size("node"))
    )

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            10.0,
            50.0,
            0.0,
            0.0,
            10.0,
            50.0,
            0.0,
            0.0,
            10.0,
            40.0,
            0.0,
            0.0,
            10.0,
            20.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_field_name_array_float_case5():
    """Topography as array, runoff rate as field name"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    runoff_rate = [
        1.0,
        1.0,
        1.0,
        1.0,
        2.0,
        2.0,
        2.0,
        2.0,
        3.0,
        3.0,
        3.0,
        3.0,
        4.0,
        4.0,
        4.0,
        4.0,
        5.0,
        5.0,
        5.0,
        5.0,
    ]

    mg.add_field("node", "topographic__elevation", topographic__elevation)
    mg.add_field("node", "runoff_rate", runoff_rate)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, "topographic__elevation", runoff_rate="runoff_rate")

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            2.0,
            16.0,
            0.0,  # KRB double checked these numbers by hand 5/15/18 - OK
            0.0,
            2.0,
            16.0,
            0.0,
            0.0,
            3.0,
            14.0,
            0.0,
            0.0,
            4.0,
            8.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_field_name_array_float_case6():
    """Topography as array, runoff rate as array"""
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    topographic__elevation = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            21.0,
            10.0,
            0.0,
            0.0,
            31.0,
            20.0,
            0.0,
            0.0,
            32.0,
            30.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    runoff_rate = [
        1.0,
        1.0,
        1.0,
        1.0,
        2.0,
        2.0,
        2.0,
        2.0,
        3.0,
        3.0,
        3.0,
        3.0,
        4.0,
        4.0,
        4.0,
        4.0,
        5.0,
        5.0,
        5.0,
        5.0,
    ]

    mg.add_field("node", "topographic__elevation", topographic__elevation)

    mg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    fa = LossyFlowAccumulator(mg, topographic__elevation, runoff_rate=runoff_rate)

    fa.run_one_step()
    reciever = np.array(
        [0, 1, 2, 3, 4, 1, 2, 7, 8, 10, 6, 11, 12, 14, 10, 15, 16, 17, 18, 19]
    )

    da = np.array(
        [
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            5.0,
            0.0,
            0.0,
            1.0,
            4.0,
            0.0,
            0.0,
            1.0,
            2.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    q = np.array(
        [
            0.0,
            2.0,
            16.0,
            0.0,  # KRB double checked these numbers by hand 5/15/18 - OK
            0.0,
            2.0,
            16.0,
            0.0,
            0.0,
            3.0,
            14.0,
            0.0,
            0.0,
            4.0,
            8.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert_array_equal(mg.at_node["flow__receiver_node"], reciever)
    assert_array_equal(mg.at_node["drainage_area"], da)
    assert_array_equal(mg.at_node["surface_water__discharge"], q)


def test_flow_accumulator_properties():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = LossyFlowAccumulator(mg)
    fa.run_one_step()

    node_drainage_area = np.array(
        [
            0.0,
            3.0,
            3.0,
            3.0,
            0.0,
            0.0,
            3.0,
            3.0,
            3.0,
            0.0,
            0.0,
            2.0,
            2.0,
            2.0,
            0.0,
            0.0,
            1.0,
            1.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    node_order_upstream = np.array(
        [
            0,
            1,
            6,
            11,
            16,
            2,
            7,
            12,
            17,
            3,
            8,
            13,
            18,
            4,
            5,
            9,
            10,
            14,
            15,
            19,
            20,
            21,
            22,
            23,
            24,
        ]
    )

    assert_array_equal(fa.node_order_upstream, node_order_upstream)
    assert_array_equal(fa.node_water_discharge, node_drainage_area)
    assert_array_equal(fa.node_drainage_area, node_drainage_area)


def test_water_discharge_in_supplied():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    mg.add_field("water__discharge_in", mg.node_x + mg.node_y, at="node")
    with pytest.deprecated_call():
        LossyFlowAccumulator(mg)


def test_bad_director_name():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director="spam")


def test_bad_director_instance():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    ld = LinearDiffuser(mg, linear_diffusivity=1.0)
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director=ld)


def test_instantiated_director_with_kwargs():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fd = FlowDirectorSteepest(mg)
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director=fd, partition_method="eggs")


def test_depression_finder_as_bad_string():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director="D8", depression_finder="spam")


def test_depression_finder_as_string():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    LossyFlowAccumulator(
        mg, flow_director="D8", depression_finder="DepressionFinderAndRouter"
    )


def test_depression_finder_as_instance():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    df = DepressionFinderAndRouter(mg)
    LossyFlowAccumulator(mg, flow_director="D8", depression_finder=df)


def test_depression_finder_bad_instance():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    ld = LinearDiffuser(mg, linear_diffusivity=1.0)
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director="D8", depression_finder=ld)


def test_instantiated_depression_finder_with_kwargs():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    df = DepressionFinderAndRouter(mg)
    with pytest.raises(ValueError):
        LossyFlowAccumulator(
            mg, flow_director="D8", depression_finder=df, routing="eggs"
        )


def test_depression_finder_bad_uninstantiated_component():
    mg = RasterModelGrid((5, 5), xy_spacing=(1, 1))
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    with pytest.raises(ValueError):
        LossyFlowAccumulator(mg, flow_director="D8", depression_finder=LinearDiffuser)


def test_hex_mfd():
    mg = HexModelGrid(5, 3)
    mg.add_field("topographic__elevation", mg.node_x + mg.node_y, at="node")
    fa = LossyFlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()



#! /usr/bin/env python
import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from landlab import RasterModelGrid
from landlab.io import read_esri_ascii, write_esri_ascii

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


def test_grid_with_no_fields(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            write_esri_ascii("test.asc", grid)


def test_grid_with_one_field(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))
    with tmpdir.as_cwd():
        files = write_esri_ascii("test.asc", grid)
        assert files == ["test.asc"]
        for fname in files:
            assert os.path.isfile(fname)


def test_grid_with_two_fields(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))
    grid.add_field("node", "land_surface__elevation", np.arange(20.0))
    with tmpdir.as_cwd():
        files = write_esri_ascii("test.asc", grid)
        files.sort()
        assert files == [
            "test_air__temperature.asc",
            "test_land_surface__elevation.asc",
        ]
        for fname in files:
            assert os.path.isfile(fname)


def test_names_keyword_as_str(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("air__temperature", np.arange(20.0), at="node")
    grid.add_field("land_surface__elevation", np.arange(20.0), at="node")

    with tmpdir.as_cwd():
        files = write_esri_ascii("test.asc", grid, names="air__temperature")
        assert files == ["test.asc"]
        assert os.path.isfile("test.asc")


def test_names_keyword_as_list(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("air__temperature", np.arange(20.0), at="node")
    grid.add_field("land_surface__elevation", np.arange(20.0), at="node")

    with tmpdir.as_cwd():
        files = write_esri_ascii("test.asc", grid, names=["air__temperature"])
        assert files == ["test.asc"]
        assert os.path.isfile("test.asc")


def test_names_keyword_multiple_names(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))
    grid.add_field("node", "land_surface__elevation", np.arange(20.0))

    with tmpdir.as_cwd():
        files = write_esri_ascii(
            "test.asc", grid, names=["air__temperature", "land_surface__elevation"]
        )
        files.sort()
        assert files == [
            "test_air__temperature.asc",
            "test_land_surface__elevation.asc",
        ]
        for fname in files:
            assert os.path.isfile(fname)


def test_names_keyword_with_bad_name(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            write_esri_ascii("test.asc", grid, names="not_a_name")


def test_clobber_keyword(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        write_esri_ascii("test.asc", grid)
        with pytest.raises(ValueError):
            write_esri_ascii("test.asc", grid)
        with pytest.raises(ValueError):
            write_esri_ascii("test.asc", grid, clobber=False)
        write_esri_ascii("test.asc", grid, clobber=True)


def test_write_then_read(tmpdir):
    grid = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0), xy_of_lower_left=(15.0, 10.0))
    grid.add_field("node", "air__temperature", np.arange(20.0))

    with tmpdir.as_cwd():
        write_esri_ascii("test.asc", grid)
        new_grid, field = read_esri_ascii("test.asc")

    assert grid.number_of_node_columns == new_grid.number_of_node_columns
    assert grid.number_of_node_rows == new_grid.number_of_node_rows
    assert grid.dx == new_grid.dx
    assert (grid.x_of_node.min(), grid.y_of_node.min()) == (15.0, 10.0)
    assert_array_almost_equal(grid.node_x, new_grid.node_x)
    assert_array_almost_equal(grid.node_y, new_grid.node_y)
    assert_array_almost_equal(field, grid.at_node["air__temperature"])



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid import raster_funcs as rfuncs


def test_with_arrays():
    """Test with arrays as arg."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))

    coords = (np.array([1.0, -1.0]), np.array([1.0, -1.0]))
    assert_array_equal(rfuncs.is_coord_on_grid(rmg, coords), np.array([True, False]))


def test_just_inside():
    """Test with points just inside the grid."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))

    assert rfuncs.is_coord_on_grid(rmg, (0.0, 4.0))
    assert rfuncs.is_coord_on_grid(rmg, (8.0 - 1e-12, 4.0))
    assert rfuncs.is_coord_on_grid(rmg, (3.0, 0.0))
    assert rfuncs.is_coord_on_grid(rmg, (3.0, 6.0 - 1e-12))


def test_just_outside():
    """Test with points just outside the grid."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))

    assert not rfuncs.is_coord_on_grid(rmg, (0.0 - 1e-12, 4.0))
    assert not rfuncs.is_coord_on_grid(rmg, (8.0, 4.0))
    assert not rfuncs.is_coord_on_grid(rmg, (3.0, 0.0 - 1e-12))
    assert not rfuncs.is_coord_on_grid(rmg, (3.0, 6.0))


def test_just_x():
    """Test check if points are within the x bounds."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2.0, 2.0))
    assert rfuncs.is_coord_on_grid(rmg, (4.0, 1.0e6), axes=(1,))
    assert not rfuncs.is_coord_on_grid(rmg, (-1.0, 1.0), axes=(1,))



#! /usr/bin/env python
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab.field import FieldError, GroupError, ModelDataFields


def test_init():
    fields = ModelDataFields()
    assert set() == fields.groups


def test_new_field_location():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    assert set(["node"]) == fields.groups


def test_add_existing_group():
    fields = ModelDataFields()
    fields.new_field_location("node", size=12)
    with pytest.raises(ValueError):
        fields.new_field_location("node", size=24)


def test_add_multiple_groups():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.new_field_location("cell", 2)
    fields.new_field_location("face", 7)
    fields.new_field_location("link", 7)
    assert set(["node", "cell", "face", "link"]) == fields.groups


def test_ones():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.new_field_location("cell", 2)

    value_array = fields.ones("node")
    assert_array_equal(np.ones(12), value_array)

    value_array = fields.ones("cell")
    assert_array_equal(np.ones(2), value_array)


def test_add_ones():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.new_field_location("cell", 2)

    fields.add_ones("z", at="node")
    assert_array_equal(np.ones(12), fields["node"]["z"])
    assert_array_equal(np.ones(12), fields.field_values("node", "z"))

    fields.add_ones("z", at="cell")
    assert_array_equal(np.ones(2), fields["cell"]["z"])
    assert_array_equal(np.ones(2), fields.field_values("cell", "z"))


def test_add_ones_return_value():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.new_field_location("cell", 2)

    rtn_value = fields.add_ones("z", at="node")
    assert_array_equal(rtn_value, np.ones(12))
    assert rtn_value is fields["node"]["z"]
    assert rtn_value is fields.field_values("node", "z")

    rtn_value = fields.add_ones("z", at="cell")
    assert_array_equal(rtn_value, np.ones(2))
    assert rtn_value is fields["cell"]["z"]
    assert rtn_value is fields.field_values("cell", "z")


def test_add_existing_field_default():
    """Test default is to not replace existing field."""
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.add_empty("z", at="node")

    with pytest.raises(FieldError):
        fields.add_empty("z", at="node")
    with pytest.raises(FieldError):
        fields.add_ones("z", at="node")
    with pytest.raises(FieldError):
        fields.add_zeros("z", at="node")


def test_add_existing_field_with_noclobber():
    """Test noclobber raises an error with an existing field."""
    fields = ModelDataFields()
    fields.new_field_location("node", 12)
    fields.add_empty("z", at="node")

    with pytest.raises(FieldError):
        fields.add_empty("z", at="node", noclobber=True)
    with pytest.raises(FieldError):
        fields.add_ones("z", at="node", noclobber=True)
    with pytest.raises(FieldError):
        fields.add_zeros("z", at="node", noclobber=True)


def test_add_field_with_noclobber():
    """Test noclobber does not raise an error with an new field."""
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    fields.add_empty("a", at="node", noclobber=True)
    assert "a" in fields["node"]

    fields.add_ones("b", at="node", noclobber=True)
    assert "b" in fields["node"]

    fields.add_zeros("c", at="node", noclobber=True)
    assert "c" in fields["node"]


def test_add_field_with_clobber():
    """Test adding a field with clobber on."""
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    assert fields.add_empty("a", at="node") is not fields.add_empty(
        "a", at="node", noclobber=False
    )
    assert fields.add_ones("b", at="node") is not fields.add_ones(
        "b", at="node", noclobber=False
    )
    assert fields.add_zeros("c", at="node") is not fields.add_zeros(
        "c", at="node", noclobber=False
    )


def test_getitem():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    assert dict() == fields["node"]
    with pytest.raises(GroupError):
        fields["cell"]
    with pytest.raises(KeyError):
        fields["cell"]


def test_at_attribute():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    assert dict() == fields.at_node
    with pytest.raises(AttributeError):
        fields.at_cell

    fields.add_ones("z", at="node")
    assert_array_equal(np.ones(12), fields.at_node["z"])


def test_has_group():
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    assert fields.has_group("node")
    assert not fields.has_group("cell")


def test_delete_field():
    fields = ModelDataFields()
    fields.new_field_location("link", 17)

    assert dict() == fields.at_link
    with pytest.raises(AttributeError):
        fields.at_node

    fields.add_zeros("link", "vals")
    assert_array_equal(np.zeros(17), fields.at_link["vals"])

    with pytest.raises(KeyError):
        fields.delete_field("node", "vals")
    fields.delete_field("link", "vals")
    with pytest.raises(KeyError):
        fields.field_units("link", "vals")
    with pytest.raises(KeyError):
        fields.at_link["vals"]


def test_scalar_field():
    """Test adding a generic scalar field."""
    fields = ModelDataFields()
    fields.new_field_location("all_over_the_place", 1)

    assert dict() == fields.at_all_over_the_place
    with pytest.raises(AttributeError):
        fields.at_cell

    fields.at_all_over_the_place["const"] = 1.0
    assert_array_equal(np.array(1.0), fields.at_all_over_the_place["const"])

    val = np.array(2.0)
    fields.at_all_over_the_place["const"] = val
    assert val is fields.at_all_over_the_place["const"]


def test_grid_field_as_array():
    """Test adding an array as a grid field."""
    fields = ModelDataFields()
    fields.new_field_location("grid", 1)

    fields.at_grid["const"] = [1.0, 2.0]
    assert_array_equal(np.array([1.0, 2.0]), fields.at_grid["const"])

    val = np.array([1.0, 2.0])
    fields.at_grid["const"] = val
    assert val is fields.at_grid["const"]

    val.shape = (1, 1, 2, 1)
    fields.at_grid["const"] = val
    assert_array_equal(np.array([1.0, 2.0]), fields.at_grid["const"])
    assert val is fields.at_grid["const"].base


def test_grid_field_add_zeros_ones_empty():
    """Test creating scalar fields with add_zeros, add_empty, and add_ones."""
    fields = ModelDataFields()
    fields.new_field_location("grid", 1)

    with pytest.raises(ValueError):
        fields.add_zeros("value", at="grid")
    with pytest.raises(ValueError):
        fields.add_empty("value", at="grid")
    with pytest.raises(ValueError):
        fields.add_ones("value", at="grid")


def test_grid_field_zeros_ones_empty():
    """Test creating scalar fields with zeros, empty, and ones."""
    fields = ModelDataFields()
    fields.new_field_location("grid", 1)
    with pytest.raises(ValueError):
        fields.zeros("grid")
    with pytest.raises(ValueError):
        fields.empty("grid")
    with pytest.raises(ValueError):
        fields.ones("grid")


def test_nd_field():
    """Test creating fields that are nd in shape."""
    fields = ModelDataFields()
    fields.new_field_location("node", 12)

    fields.add_field("new_value", np.ones((12, 4, 5)), at="node")
    fields.add_field("newer_value", np.ones((12, 4)), at="node")

    with pytest.raises(ValueError):
        fields.add_field("newest_value", np.ones((13, 4, 5)), at="node")
    with pytest.raises(ValueError):
        fields.add_field("newestest_value", np.ones((13)), at="node")



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.io.netcdf import write_raster_netcdf

try:
    import netCDF4 as nc
except ImportError:
    pass


def test_append_with_time(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", np.ones(12, dtype=np.int64))

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, append=False, format="NETCDF4", time=0)
        field.at_node["topographic__elevation"] *= 2
        write_raster_netcdf("test.nc", field, append=True, format="NETCDF4", time=1.0)

        root = nc.Dataset("test.nc", "r", format="NETCDF4")

        for name in ["topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:],
                [
                    [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                    [[2, 2, 2], [2, 2, 2], [2, 2, 2], [2, 2, 2]],
                ],
            )
            assert root.variables[name][:].dtype == "int64"
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 2

        assert "t" in root.variables
        assert_array_equal(root.variables["t"][:], [0.0, 1.0])

        root.close()


def test_without_time(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", np.ones(12, dtype=np.int64))

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, append=False, format="NETCDF4")

        root = nc.Dataset("test.nc", "r", format="NETCDF4")

        for name in ["topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:], [[[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]]
            )
            assert root.variables[name][:].dtype == "int64"
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 1

        assert "t" not in root.variables

        root.close()


def test_with_time(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", np.ones(12, dtype=np.int64))

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, append=False, format="NETCDF4", time=0.0)

        root = nc.Dataset("test.nc", "r", format="NETCDF4")

        for name in ["topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:], [[[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]]]
            )
            assert root.variables[name][:].dtype == "int64"
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 1

        assert "t" in root.variables
        assert_array_equal(root.variables["t"][:], [0.0])

        root.close()


def test_with_time_netcdf3(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", 2.0 * np.arange(12.0))
    field.add_field("node", "uplift_rate", 2.0 * np.arange(12.0))

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, format="NETCDF3_64BIT", time=10.0)

        root = nc.Dataset("test.nc", "r", format="NETCDF3_64BIT")

        for name in ["uplift_rate", "topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:],
                [
                    [
                        [0.0, 2.0, 4.0],
                        [6.0, 8.0, 10.0],
                        [12.0, 14.0, 16.0],
                        [18.0, 20.0, 22.0],
                    ]
                ],
            )
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 1

        assert "t" in root.variables
        assert_array_equal(root.variables["t"][:], [10.0])

        root.close()


def test_append_with_time_netcdf3(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("topographic__elevation", np.ones(12), at="node")

    with tmpdir.as_cwd():
        write_raster_netcdf(
            "test.nc", field, append=False, format="NETCDF3_64BIT", time=0
        )
        field.at_node["topographic__elevation"] *= 2
        write_raster_netcdf(
            "test.nc", field, append=True, format="NETCDF3_64BIT", time=1.0
        )

        root = nc.Dataset("test.nc", "r", format="NETCDF3_64BIT")

        for name in ["topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:],
                [
                    [[1, 1, 1], [1, 1, 1], [1, 1, 1], [1, 1, 1]],
                    [[2, 2, 2], [2, 2, 2], [2, 2, 2], [2, 2, 2]],
                ],
            )
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 2

        assert "t" in root.variables
        assert_array_equal(root.variables["t"][:], [0.0, 1.0])

        root.close()


def test_append_without_time_netcdf3(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("topographic__elevation", np.ones(12), at="node")

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, append=False, format="NETCDF3_64BIT")
        field.at_node["topographic__elevation"] *= 2
        write_raster_netcdf("test.nc", field, append=True, format="NETCDF3_64BIT")

        root = nc.Dataset("test.nc", "r", format="NETCDF3_64BIT")

        for name in ["topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:], [[[2, 2, 2], [2, 2, 2], [2, 2, 2], [2, 2, 2]]]
            )
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 1

        assert "t" not in root.variables

        root.close()


def test_without_time_netcdf3(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", 2.0 * np.arange(12.0))
    field.add_field("node", "uplift_rate", 2.0 * np.arange(12.0))

    with tmpdir.as_cwd():
        write_raster_netcdf("test.nc", field, format="NETCDF3_64BIT")

        root = nc.Dataset("test.nc", "r", format="NETCDF3_64BIT")

        for name in ["uplift_rate", "topographic__elevation"]:
            assert name in root.variables
            assert_array_equal(
                root.variables[name][:],
                [
                    [
                        [0.0, 2.0, 4.0],
                        [6.0, 8.0, 10.0],
                        [12.0, 14.0, 16.0],
                        [18.0, 20.0, 22.0],
                    ]
                ],
            )
            assert "nt" in root.dimensions
            assert len(root.dimensions["nt"]) == 1

        assert "t" not in root.variables

        root.close()


def test_names_keyword(tmpdir):
    field = RasterModelGrid((4, 3))
    field.add_field("node", "topographic__elevation", np.arange(12.0))
    field.add_field("node", "uplift_rate", 2.0 * np.arange(12.0))

    with tmpdir.as_cwd():
        write_raster_netcdf(
            "test.nc", field, format="NETCDF3_64BIT", names="uplift_rate"
        )

        root = nc.Dataset("test.nc", "r", format="NETCDF3_64BIT")

        assert "topographic__elevation" not in root.variables
        assert "uplift_rate" in root.variables

        assert_array_equal(
            root.variables["uplift_rate"][:],
            [
                [
                    [0.0, 2.0, 4.0],
                    [6.0, 8.0, 10.0],
                    [12.0, 14.0, 16.0],
                    [18.0, 20.0, 22.0],
                ]
            ],
        )
        assert "nt" in root.dimensions
        assert len(root.dimensions["nt"]) == 1

        root.close()



import matplotlib.pyplot as plt

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator
from landlab.plot.drainage_plot import drainage_plot


def make_grid():
    mg = RasterModelGrid((10, 10), xy_spacing=(1, 1))
    mg.add_field(
        "topographic__elevation", mg.node_x ** 2 + mg.node_y ** 2 + mg.node_y, at="node"
    )
    return mg


def test_steepest():
    mg = make_grid()
    fa = FlowAccumulator(mg)
    fa.run_one_step()
    plt.figure()
    drainage_plot(mg)


def test_mfd():
    mg = make_grid()
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()
    plt.figure()
    drainage_plot(mg)



"""Test the storm generator and simple stream power eroder both execute."""
import os

import numpy as np

from landlab import ModelParameterDictionary, RasterModelGrid
from landlab.components import (
    FlowAccumulator,
    PrecipitationDistribution,
    StreamPowerEroder,
)

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_storms():
    input_file_string = os.path.join(_THIS_DIR, "drive_sp_params_storms.txt")
    inputs = ModelParameterDictionary(input_file_string, auto_type=True)
    nrows = inputs.read_int("nrows")
    ncols = inputs.read_int("ncols")
    dx = inputs.read_float("dx")
    dt = inputs.read_float("dt")
    uplift = inputs.read_float("uplift_rate")

    mean_duration = inputs.read_float("mean_storm")
    mean_interstorm = inputs.read_float("mean_interstorm")
    mean_depth = inputs.read_float("mean_depth")

    storm_run_time = inputs.read_float("storm_run_time")
    delta_t = inputs.read_float("delta_t")
    mg = RasterModelGrid((nrows, ncols), xy_spacing=dx)

    mg.add_zeros("topographic__elevation", at="node")
    z = mg.zeros(at="node")
    mg["node"]["topographic__elevation"] = z + np.random.rand(len(z)) / 1000.0
    mg.add_zeros("water__unit_flux_in", at="node")

    precip = PrecipitationDistribution(
        mean_storm_duration=mean_duration,
        mean_interstorm_duration=mean_interstorm,
        mean_storm_depth=mean_depth,
        total_t=storm_run_time,
        delta_t=delta_t,
    )
    fr = FlowAccumulator(mg, flow_director="D8")
    sp = StreamPowerEroder(mg, **inputs)

    for (
        interval_duration,
        rainfall_rate,
    ) in precip.yield_storm_interstorm_duration_intensity():
        if rainfall_rate != 0.0:
            mg.at_node["water__unit_flux_in"].fill(rainfall_rate)
            fr.run_one_step()
            sp.run_one_step(dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += (
            uplift * interval_duration
        )



# -*- coding: utf-8 -*-
"""
GroundwaterDupuitPercolator Component

@author: G Tucker
"""

import numpy as np

from landlab import Component
from landlab.grid.mappers import map_mean_of_link_nodes_to_link
from landlab.utils import return_array_at_link, return_array_at_node

ACTIVE_LINK = 0


class GroundwaterDupuitPercolator(Component):
    """
    Simulate groundwater flow in a shallow unconfined aquifer.

    The GroundwaterDupuitPercolator uses the Dupuit approximation that the
    hydraulic gradient is equal to the slope of the water table.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((3, 3))
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> gdp = GroundwaterDupuitPercolator(mg)

    Notes
    -----
    Groundwater discharge per unit length, q, is calculated as:

        q = - K H dw/dx,

    where K is hydraulic conductivity, H is aquifer thickness, w is water table
    height, and x is horizontal distance.

    An explicit forward-in-time finite-volume method is used to implement a
    numerical solution. Flow discharge between neighboring nodes is calculated
    using the average depth at the nodes.

    Note that the current version does NOT handle surface seepage.
    """

    _name = "GroundwaterDupuitPercolator"

    _input_var_names = set(("topographic__elevation", "aquifer_base__elevation"))

    _output_var_names = set(
        (
            "aquifer__thickness",
            "water_table__elevation",
            "hydraulic__gradient",
            "groundwater__specific_discharge",
            "groundwater__velocity",
        )
    )

    _var_units = {
        "topographic__elevation": "m",
        "aquifer_base__elevation": "m",
        "aquifer__thickness": "m",
        "water_table__elevation": "m",
        "hydraulic__gradient": "m/m",
        "groundwater__specific_discharge": "m2/s",
        "groundwater__velocity": "m/s",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "aquifer_base__elevation": "node",
        "aquifer__thickness": "node",
        "water_table__elevation": "node",
        "hydraulic__gradient": "link",
        "groundwater__specific_discharge": "link",
        "groundwater__velocity": "link",
    }

    _var_doc = {
        "topographic__elevation": "elevation of land surface",
        "aquifer_base__elevation": "elevation of impervious layer",
        "aquifer__thickness": "thickness of saturated zone",
        "water_table__elevation": "elevation of water table",
        "hydraulic__gradient": "gradient of water table in link direction",
        "groundwater__specific_discharge": "discharge per width in link dir",
        "groundwater__velocity": "velocity of groundwater in link direction",
    }

    def __init__(self, grid, hydraulic_conductivity=0.01, recharge_rate=1.0e-8):
        """Initialize the GroundwaterDupuitPercolator.

        Parameters
        ----------
        grid: ModelGrid
            Landlab ModelGrid object
        hydraulic_conductivity: float, field name (link), or array of float
            saturated hydraulic conductivity, m/s
            Default = 0.01 m/s
        recharge_rate: float, field name (node), or array of float
            Rate of recharge, m/s
            Default = 1.0e-8 m/s
        """
        # Store grid
        self._grid = grid

        # Shorthand
        self._cores = grid.core_nodes
        self._inactive_links = np.where(grid.status_at_link != ACTIVE_LINK)[0]

        # Convert parameters to fields if needed, and store a reference
        self._K = return_array_at_link(grid, hydraulic_conductivity)
        self._recharge = return_array_at_node(grid, recharge_rate)

        # Create fields:

        if "topographic__elevation" in self._grid.at_node:
            self._elev = self._grid.at_node["topographic__elevation"]
        else:
            self._elev = self._grid.add_ones("node", "topographic__elevation")

        if "aquifer_base__elevation" in self._grid.at_node:
            self._base = self._grid.at_node["aquifer_base__elevation"]
        else:
            self._base = self._grid.add_zeros("node", "aquifer_base__elevation")

        if "water_table__elevation" in self._grid.at_node:
            self._wtable = self._grid.at_node["water_table__elevation"]
        else:
            self._wtable = self._grid.add_zeros("node", "water_table__elevation")

        if "aquifer__thickness" in self._grid.at_node:
            self._thickness = self._grid.at_node["aquifer__thickness"]
        else:
            self._thickness = self._grid.add_zeros("node", "aquifer__thickness")
            self._thickness[:] = self._wtable - self._base

        if "hydraulic__gradient" in self._grid.at_link:
            self._hydr_grad = self._grid.at_link["hydraulic__gradient"]
        else:
            self._hydr_grad = self._grid.add_zeros("link", "hydraulic__gradient")

        if "groundwater__specific_discharge" in self._grid.at_link:
            self._q = self._grid.at_link["groundwater__specific_discharge"]
        else:
            self._q = self._grid.add_zeros("link", "groundwater__specific_discharge")

        if "groundwater__velocity" in self._grid.at_link:
            self._vel = self._grid.at_link["groundwater__velocity"]
        else:
            self._vel = self._grid.add_zeros("link", "groundwater__velocity")

    def run_one_step(self, dt, **kwds):
        """
        Advance component by one time step of size dt.

        Parameters
        ----------
        dt: float (time in seconds)
            The imposed timestep.
        """

        # Calculate hydraulic gradient
        self._hydr_grad[:] = self._grid.calc_grad_at_link(self._wtable)
        self._hydr_grad[self._inactive_links] = 0.0

        # Calculate groundwater velocity
        self._vel[:] = -self._K * self._hydr_grad
        self._vel[self._grid.status_at_link != 0] = 0.0

        # Aquifer thickness at links
        hlink = map_mean_of_link_nodes_to_link(self._grid, "aquifer__thickness")

        # Calculate specific discharge
        self._q[:] = hlink * self._vel

        # Mass balance
        dqdx = self._grid.calc_flux_div_at_node(self._q)
        dhdt = self._recharge - dqdx

        # Update
        self._thickness[self._grid.core_nodes] += dhdt[self._cores] * dt

        # Recalculate water surface height
        self._wtable[self._grid.core_nodes] = (
            self._base[self._cores] + self._thickness[self._cores]
        )



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid.raster_gradients import calc_grad_across_cell_corners


def test_scalar_arg():
    """Test using a scalar for cell arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    grads = calc_grad_across_cell_corners(rmg, values_at_nodes, 0)
    assert_array_equal(grads, np.array([[6.0, 4.0, -6.0, -4.0]]) / np.sqrt(2.0))


def test_iterable():
    """Test using an iterable for cell arg."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    grads = rmg.calc_grad_across_cell_corners(values_at_nodes, [0, 4])
    assert_array_equal(
        grads, np.array([[6.0, 4.0, -6.0, -4.0], [6.0, 4.0, -6.0, -4.0]]) / np.sqrt(2.0)
    )


def test_with_no_cell_id_arg():
    """Test without using an arg for cell id."""
    rmg = RasterModelGrid((4, 5))
    values = np.array(
        [0, 1, 3, 6, 10, 0, 1, 3, 6, 10, 0, 1, 3, 5, 10, 0, 1, -3, 6, 10], dtype=float
    )
    grads = rmg.calc_grad_across_cell_corners(values)

    assert_array_equal(
        grads,
        (1.0 / np.sqrt(2.0))
        * np.array(
            [
                [2.0, -1.0, -1.0, 2.0],
                [2.0, -2.0, -2.0, 3.0],
                [4.0, -3.0, -3.0, 4.0],
                [-4.0, -1.0, -1.0, 2.0],
                [3.0, -2.0, -2.0, 3.0],
                [5.0, -8.0, -2.0, 5.0],
            ]
        ),
    )


def test_with_out_keyword():
    """Test with out keyword."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    out = np.empty((1, 4))
    rtn = rmg.calc_grad_across_cell_corners(values_at_nodes, 5, out=out)
    assert rtn is out
    assert_array_equal(out, np.array([[6.0, 4.0, -6.0, -4.0]]) / np.sqrt(2))



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid

ELEMENTS = ["node", "cell", "link", "face"]
# ELEMENTS += ['core_node', 'core_cell', 'active_link', 'active_face']
TYPES = ["float", "int", "bool"]


def generate_zeros_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                assert_array_equal(
                    rmg.zeros(centering=element),
                    np.zeros(number_of_elements, dtype=np.float),
                )

            _test.description = "%s.test_zeros_%s_%s" % (__name__, type, element)
            yield _test


def generate_add_zeros_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_zeros(element, "name")
                assert rtn_values is rmg.field_values(element, "name")
                assert_array_equal(
                    rtn_values, np.zeros(number_of_elements, dtype=np.float)
                )

            _test.description = "%s.test_add_zeros_%s_%s" % (__name__, type, element)
            yield _test


def generate_ones_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                assert_array_equal(
                    rmg.ones(centering=element),
                    np.ones(number_of_elements, dtype=np.float),
                )

            _test.description = "%s.test_zeros_%s_%s" % (__name__, type, element)
            yield _test


def generate_add_ones_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_ones(element, "name")
                assert rtn_values is rmg.field_values(element, "name")
                assert_array_equal(
                    rtn_values, np.ones(number_of_elements, dtype=np.float)
                )

            _test.description = "%s.test_add_zeros_%s_%s" % (__name__, type, element)
            yield _test


def generate_empty_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                assert rmg.empty(centering=element).size == number_of_elements

            _test.description = "%s.test_zeros_%s_%s" % (__name__, type, element)
            yield _test


def generate_add_empty_tests():
    for element in ELEMENTS:
        for type in TYPES:

            def _test():
                rmg = RasterModelGrid((4, 5))
                number_of_elements = rmg.number_of_elements(element)
                rtn_values = rmg.add_empty(element, "name")
                assert rtn_values is rmg.field_values(element, "name")
                assert_array_equal(rtn_values.size, number_of_elements)

            _test.description = "%s.test_zeros_%s_%s" % (__name__, type, element)
            yield _test



from landlab.plot.channel_profile import analyze_channel_network_and_plot
from landlab.plot.imshow import imshow_grid, imshow_grid_at_node

__all__ = ["imshow_grid", "imshow_grid_at_node", "analyze_channel_network_and_plot"]



import numpy as np
import pytest

from landlab import FieldError, RasterModelGrid
from landlab.utils.return_array import return_array_at_link, return_array_at_node


def test_no_field():
    mg = RasterModelGrid((10, 10))
    with pytest.raises(FieldError):
        return_array_at_node(mg, "spam")
    with pytest.raises(FieldError):
        return_array_at_link(mg, "spam")


def test_return_array():
    mg = RasterModelGrid((10, 10))
    node_vals = np.arange(mg.number_of_nodes)
    out = return_array_at_node(mg, node_vals)

    np.testing.assert_array_equal(np.arange(mg.number_of_nodes), out)

    link_vals = np.arange(mg.number_of_links)
    out = return_array_at_link(mg, link_vals)

    np.testing.assert_array_equal(np.arange(mg.number_of_links), out)



import pytest

from landlab import NetworkModelGrid, RasterModelGrid
from landlab.values.synthetic import _STATUS


@pytest.fixture
def four_by_four_raster():
    mg = RasterModelGrid((4, 4))
    return mg


@pytest.fixture
def simple_network():
    y_of_node = (0, 1, 2, 2)
    x_of_node = (0, 0, -1, 1)
    nodes_at_link = ((1, 0), (2, 1), (3, 1))
    mg = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)
    return mg


def pytest_generate_tests(metafunc):
    if "at" in metafunc.fixturenames:
        metafunc.parametrize("at", ("node", "link", "patch", "corner", "face", "cell"))
    if "node_bc" in metafunc.fixturenames:
        metafunc.parametrize("node_bc", list(_STATUS["node"].keys()))
    if "link_bc" in metafunc.fixturenames:
        metafunc.parametrize("link_bc", list(_STATUS["link"].keys()))



import numpy as np
from numpy.testing import assert_array_equal

from landlab import CLOSED_BOUNDARY as CB, FIXED_VALUE_BOUNDARY as FV, RasterModelGrid


def test_get_status():
    """Test getting the node status array."""
    grid = RasterModelGrid((4, 5))
    assert_array_equal(
        grid.status_at_node,
        [FV, FV, FV, FV, FV, FV, 0, 0, 0, FV, FV, 0, 0, 0, FV, FV, FV, FV, FV, FV],
    )


def test_set_status_with_scalar():
    """Test setting status with a scalar."""
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[6] = 2

    assert_array_equal(
        grid.status_at_node,
        [FV, FV, FV, FV, FV, FV, 2, 0, 0, FV, FV, 0, 0, 0, FV, FV, FV, FV, FV, FV],
    )


def test_set_status_with_slice():
    """Test setting status with an array slice."""
    grid = RasterModelGrid((4, 5))
    grid.status_at_node[6:9] = 2

    assert_array_equal(
        grid.status_at_node,
        [FV, FV, FV, FV, FV, FV, 2, 2, 2, FV, FV, 0, 0, 0, FV, FV, FV, FV, FV, FV],
    )


def test_set_status_with_array_bool():
    """Test setting node status with boolean array."""
    grid = RasterModelGrid((4, 5))
    inds = np.full((20,), False, dtype=bool)
    inds[6] = True
    inds[7] = True
    inds[13] = True

    grid.status_at_node[inds] = 2
    assert_array_equal(
        grid.status_at_node,
        [FV, FV, FV, FV, FV, FV, 2, 2, 0, FV, FV, 0, 0, 2, FV, FV, FV, FV, FV, FV],
    )


def test_set_with_itemset():
    grid = RasterModelGrid((4, 5))
    grid.status_at_node.itemset(7, 2)

    assert_array_equal(
        grid.status_at_node,
        [FV, FV, FV, FV, FV, FV, 0, 2, 0, FV, FV, 0, 0, 0, FV, FV, FV, FV, FV, FV],
    )


def test_set_status_with_array():
    """Test that active links are reset after changing the node status."""
    grid = RasterModelGrid((4, 5))

    assert_array_equal(
        grid.active_links,
        [5, 6, 7, 9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 23, 24, 25],
    )

    grid.status_at_node[:5] = CB
    assert_array_equal(
        grid.status_at_node,
        [CB, CB, CB, CB, CB, FV, 0, 0, 0, FV, FV, 0, 0, 0, FV, FV, FV, FV, FV, FV],
    )

    assert_array_equal(
        grid.active_links, [9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 23, 24, 25]
    )



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlowBates

last updated: 3/14/16
"""
import numpy as np

from landlab.components.overland_flow import OverlandFlowBates

(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_Bates_name(bates):
    assert bates.name == "OverlandFlowBates"


def test_Bates_input_var_names(bates):
    assert set(bates.input_var_names) == set(
        ("surface_water__depth", "topographic__elevation")
    )


def test_Bates_output_var_names(bates):
    assert set(bates.output_var_names) == set(
        ("surface_water__depth", "surface_water__discharge", "water_surface__gradient")
    )


def test_Bates_var_units(bates):
    assert set(bates.input_var_names) | set(bates.output_var_names) == set(
        dict(bates.units).keys()
    )

    assert bates.var_units("surface_water__depth") == "m"
    assert bates.var_units("surface_water__discharge") == "m3/s"
    assert bates.var_units("water_surface__gradient") == "m/m"
    assert bates.var_units("topographic__elevation") == "m"


def test_field_initialized_to_zero(bates):
    for name in bates.grid["node"].keys():
        field = bates.grid["node"][name]
        if name != "surface_water__depth":
            assert np.all(np.isclose(field, 0.0))
        else:
            assert np.all(np.isclose(field, 0.001))
            # this remains broken, and needs JA's attention


def test_grid_shape(bates):
    assert bates.grid.number_of_node_rows == _SHAPE[0]
    assert bates.grid.number_of_node_columns == _SHAPE[1]


def test_Bates_analytical():
    from landlab import RasterModelGrid

    grid = RasterModelGrid((32, 240), xy_spacing=25)
    grid.add_zeros("node", "surface_water__depth")
    grid.add_zeros("node", "topographic__elevation")
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    bates = OverlandFlowBates(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0
    bates.dt = 1.0
    while time < 500:
        bates.overland_flow(grid)
        h_boundary = ((7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 3) * time) ** (3.0 / 7.0)
        grid.at_node["surface_water__depth"][grid.nodes[1:-1, 1]] = h_boundary
        time += bates.dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = -(7.0 / 3.0) * (0.01 ** 2) * (0.4 ** 2) * (x - (0.4 * 500))

    h_analytical[np.where(h_analytical > 0)] = h_analytical[
        np.where(h_analytical > 0)
    ] ** (3.0 / 7.0)
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hBates = bates.h.reshape(grid.shape)
    hBates = hBates[1][1:]
    hBates = np.append(hBates, [0])
    np.testing.assert_almost_equal(h_analytical, hBates, decimal=1)



# -*- coding: utf-8 -*-
"""
This is an implementation of Vaughan Voller's experimental boundary method
reduced complexity flow router. Credit: Voller, Hobley, Paola.

Created on Fri Feb 20 09:32:27 2015

@author: danhobley (SiccarPoint), after volle001@umn.edu
"""

# ##in the diagonal case, even closed edges can produce "drag". Is this right?
# Could suppress by mirroring the diagonals

import inspect

import numpy as np

from landlab import (
    ACTIVE_LINK,
    FIXED_LINK,
    INACTIVE_LINK,
    Component,
    FieldError,
    RasterModelGrid,
)
from landlab.utils.decorators import use_file_name_or_kwds


class PotentialityFlowRouter(Component):
    """
    This class implements Voller, Hobley, and Paola's experimental matrix
    solutions for flow routing.
    Options are permitted to allow "abstract" routing (flow enforced downslope,
    but no particular assumptions are made about the governing equations), or
    routing according to the Chezy or Manning equations. This routine assumes
    that water is distributed evenly over the surface of the cell in deriving
    the depth, and does not assume channelization. You will need to back-
    calculate channel depths for yourself using known widths at each node
    if that is what you want.

    It is VITAL you initialize this component AFTER setting boundary
    conditions.

    Note
    ----
    This is a "research grade" component, and is subject to dramatic change
    with little warning. No guarantees are made regarding its accuracy or
    utility. It is not recommended for user use yet!
    """

    _name = "PotentialityFlowRouter"

    _input_var_names = ("topographic__elevation", "water__unit_flux_in")

    _output_var_names = (
        "surface_water__discharge",
        "water__discharge_x_component",
        "water__discharge_y_component",
        "flow__potential",
        "surface_water__depth",
    )

    _var_units = {
        "topographic__elevation": "m",
        "water__unit_flux_in": "m/s",
        "surface_water__discharge": "m**3/s",
        "water__discharge_x_component": "m**3/s",
        "water__discharge_y_component": "m**3/s",
        "flow__potential": "m**3/s",
        "surface_water__depth": "m",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "water__unit_flux_in": "node",
        "surface_water__discharge": "node",
        "water__discharge_x_component": "node",
        "water__discharge_y_component": "node",
        "flow__potential": "node",
        "surface_water__depth": "node",
    }

    _var_doc = {
        "topographic__elevation": "Land surface topographic elevation",
        "water__unit_flux_in": "External volume water per area per time input to each node (e.g., rainfall rate)",
        "surface_water__discharge": "Magnitude of volumetric water flux through each node",
        "water__discharge_x_component": "x component of resolved water flux through node",
        "water__discharge_y_component": "y component of resolved water flux through node",
        "flow__potential": 'Value of the hypothetical field "K", used to force water flux to flow downhill',
        "surface_water__depth": "If Manning or Chezy specified, the depth of flow in the cell, calculated assuming flow occurs over the whole surface",
    }

    _min_slope_thresh = (
        1.0e-24
    )  # if your flow isn't connecting up, this probably needs to be reduced

    @use_file_name_or_kwds
    def __init__(
        self,
        grid,
        method="D8",
        flow_equation="default",
        Chezys_C=None,
        Mannings_n=0.03,
        return_components=False,
        suppress_closed_node_friction=True,
        **kwds
    ):
        """Initialize flow router.

        Parameters
        ----------
        grid : ModelGrid
            A grid
        params : dict
            Input parameters. Optional parameters are:

            *  `flow_equation` : options are ``default`` (default),
               ``Manning``, or ``Chezy``.  If Equation is ``Manning`` or
               ``Chezy``, you must also specify:

               *  ``Mannings_n`` (if ``Manning``) : float
               *  ``Chezys_C`` (if ``Chezy``) : float
        """
        if RasterModelGrid in inspect.getmro(grid.__class__):
            assert grid.number_of_node_rows >= 3
            assert grid.number_of_node_columns >= 3
            self._raster = True
        else:
            self._raster = False
        assert type(return_components) is bool

        self._grid = grid
        self.equation = flow_equation

        if method not in ("D8", "D4"):
            raise ValueError("'method' must be one of 'D8', 'D4'")
        if self.equation not in ("default", "Chezy", "Manning"):
            raise ValueError("'equation' must be one of 'default', 'Chezy', 'Manning'")
        if self.equation == "Chezy" and Chezys_C is None:
            raise ValueError("missing value for Chezys_C")

        if self.equation == "Chezy":
            self.chezy_C = Chezys_C
        elif self.equation == "Manning":
            self.manning_n = Mannings_n
        if method == "D8":
            self.route_on_diagonals = True
        else:
            self.route_on_diagonals = False
        self.return_components = return_components
        if self.return_components:
            raise ValueError(
                "Component cannot presently return discharge in x and y "
                + "directions. Consider using the discharges_at_links "
                + "property to do the mapping yourself."
            )
        self.suppress_closed_node_friction = suppress_closed_node_friction

        # hacky fix because water__discharge is defined on both links and nodes
        for out_field in self._output_var_names:
            if self._var_mapping[out_field] == "node":
                try:
                    self.grid.add_zeros(
                        self._var_mapping[out_field], out_field, dtype=float
                    )
                except FieldError:
                    pass
            else:
                pass
            try:
                self.grid.add_zeros("node", "surface_water__discharge", dtype=float)
            except FieldError:
                pass

        if self._raster:
            self.equiv_circ_diam = 2.0 * np.sqrt(grid.dx * grid.dy / np.pi)
        else:
            for_cell_areas = 2.0 * np.sqrt(grid.area_of_cell / np.pi)
            mean_A = for_cell_areas.mean()
            self.equiv_circ_diam = for_cell_areas[grid.cell_at_node]
            self.equiv_circ_diam[grid.cell_at_node == -1] = mean_A
        # ^this is the equivalent seen CSWidth of a cell for a flow in a
        # generic 360 direction
        if self.route_on_diagonals and self._raster:
            self._discharges_at_link = np.empty(grid.number_of_d8)
        else:
            self._discharges_at_link = self.grid.empty("link")

    def route_flow(self, **kwds):
        """
        """
        grid = self.grid
        self._K = grid.at_node["flow__potential"]
        self._Qw = grid.at_node["surface_water__discharge"]
        z = grid.at_node["topographic__elevation"]
        qwater_in = grid.at_node["water__unit_flux_in"].copy()
        qwater_in[grid.node_at_cell] *= grid.area_of_cell
        prev_K = self._K.copy()
        mismatch = 10000.0
        # do the ortho nodes first, in isolation
        g = grid.calc_grad_at_link(z)
        if self.equation != "default":
            g = np.sign(g) * np.sqrt(np.fabs(g))
            # ^...because both Manning and Chezy actually follow sqrt
            # slope, not slope
        # weight by face width - NO, because diags
        # g *= grid.width_of_face[grid.face_at_link]
        link_grad_at_node_w_dir = g[grid.links_at_node] * grid.active_link_dirs_at_node
        # active_link_dirs strips "wrong" face widths

        # now outgoing link grad sum
        outgoing_sum = (
            np.sum((link_grad_at_node_w_dir).clip(0.0), axis=1) + self._min_slope_thresh
        )
        pos_incoming_link_grads = (-link_grad_at_node_w_dir).clip(0.0)

        if not self.route_on_diagonals or not self._raster:
            while mismatch > 1.0e-6:
                K_link_ends = self._K[grid.adjacent_nodes_at_node]
                incoming_K_sum = (pos_incoming_link_grads * K_link_ends).sum(
                    axis=1
                ) + self._min_slope_thresh
                self._K[:] = (incoming_K_sum + qwater_in) / outgoing_sum
                mismatch = np.sum(np.square(self._K - prev_K))
                prev_K = self._K.copy()

            upwind_K = grid.map_value_at_max_node_to_link(z, self._K)
            self._discharges_at_link[:] = upwind_K * g
            self._discharges_at_link[grid.status_at_link == INACTIVE_LINK] = 0.0
        else:
            active_diagonal_dirs_at_node = self.grid.diagonal_dirs_at_node * (
                self.grid.diagonal_status_at_node == ACTIVE_LINK
            )

            # grad on diags:
            gwd = np.empty(grid.number_of_d8, dtype=float)
            gd = gwd[grid.number_of_links :]
            gd[:] = z[grid.nodes_at_diagonal[:, 1]] - z[grid.nodes_at_diagonal[:, 0]]
            gd /= grid._length_of_d8[grid.number_of_links :]
            if self.equation != "default":
                gd[:] = np.sign(gd) * np.sqrt(np.fabs(gd))
            diag_grad_at_node_w_dir = (
                gwd[grid.d8s_at_node[:, 4:]] * active_diagonal_dirs_at_node
            )

            outgoing_sum += np.sum(diag_grad_at_node_w_dir.clip(0.0), axis=1)
            pos_incoming_diag_grads = (-diag_grad_at_node_w_dir).clip(0.0)
            while mismatch > 1.0e-6:
                K_link_ends = self._K[grid.adjacent_nodes_at_node]
                K_diag_ends = self._K[grid.diagonal_adjacent_nodes_at_node]
                incoming_K_sum = (
                    (pos_incoming_link_grads * K_link_ends).sum(axis=1)
                    + (pos_incoming_diag_grads * K_diag_ends).sum(axis=1)
                    + self._min_slope_thresh
                )
                self._K[:] = (incoming_K_sum + qwater_in) / outgoing_sum
                mismatch = np.sum(np.square(self._K - prev_K))
                prev_K = self._K.copy()

            # ^this is necessary to suppress stupid apparent link Qs at flow
            # edges, if present.
            upwind_K = grid.map_value_at_max_node_to_link(z, self._K)
            upwind_diag_K = np.where(
                z[grid.nodes_at_diagonal[:, 1]] > z[grid.nodes_at_diagonal[:, 0]],
                self._K[grid.nodes_at_diagonal[:, 1]],
                self._K[grid.nodes_at_diagonal[:, 0]],
            )
            self._discharges_at_link[: grid.number_of_links] = upwind_K * g
            self._discharges_at_link[grid.number_of_links :] = upwind_diag_K * gd
            self._discharges_at_link[grid.status_at_d8 == FIXED_LINK] = 0.0

        np.multiply(self._K, outgoing_sum, out=self._Qw)
        # there is no sensible way to save discharges at links, if we route
        # on diagonals.
        # for now, let's make a property

        if self.return_components:
            pass
            # can't do this yet

        # now process uval and vval to give the depths, if Chezy or Manning:
        if self.equation == "Chezy":
            # Chezy: Q = C*Area*sqrt(depth*slope)
            grid.at_node["surface_water__depth"][:] = (
                grid.at_node["flow__potential"] / self.chezy_C / self.equiv_circ_diam
            ) ** (2.0 / 3.0)
        elif self.equation == "Manning":
            # Manning: Q = w/n*depth**(5/3)
            grid.at_node["surface_water__depth"][:] = (
                grid.at_node["flow__potential"] * self.manning_n / self.equiv_circ_diam
            ) ** 0.6
        else:
            pass

    def run_one_step(self, **kwds):
        self.route_flow(**kwds)

    @property
    def discharges_at_links(self):
        """Return the discharges at links.

        Note that if diagonal routing, this will return number_of_d8.
        Otherwise, it will be number_of_links.
        """
        return self._discharges_at_link



import matplotlib.pyplot as plt
import numpy as np


def plot_nodes(graph, color="r", with_id=True, markersize=10):
    for node in range(len(graph.x_of_node)):
        x, y = graph.x_of_node[node], graph.y_of_node[node]
        plt.plot(
            graph.x_of_node[node],
            graph.y_of_node[node],
            "o",
            color=color,
            markersize=markersize,
        )
        if with_id:
            plt.text(x, y, node, color=color, size=16)


def plot_links(
    graph, color="b", linestyle="solid", with_id=True, as_arrow=True, linewidth=None
):
    if as_arrow:
        head_width = 0.1
    else:
        head_width = 0.0
    for link, nodes in enumerate(graph.nodes_at_link):
        x, y = graph.x_of_node[nodes[0]], graph.y_of_node[nodes[0]]
        dx, dy = graph.x_of_node[nodes[1]] - x, graph.y_of_node[nodes[1]] - y
        plt.arrow(
            x,
            y,
            dx,
            dy,
            head_width=head_width,
            linewidth=linewidth,
            length_includes_head=True,
            color=color,
            linestyle=linestyle,
        )
        if with_id:
            plt.text(x + dx * 0.5, y + dy * 0.5, link, size=16, color=color)


def plot_patches(graph, color="g"):
    for patch, nodes in enumerate(graph.nodes_at_patch):
        x, y = np.mean(graph.x_of_node[nodes]), np.mean(graph.y_of_node[nodes])
        plt.text(x, y, patch, color=color, size=16)


def plot_graph(graph, at="node,link,patch", with_id=True):
    locs = [loc.strip() for loc in at.split(",")]
    for loc in locs:
        if loc not in ("node", "link", "patch", "corner", "face", "cell"):
            raise ValueError('{at}: "at" element not understood'.format(at=loc))

    plt.plot(graph.x_of_node, graph.y_of_node, ".", color="r")
    plt.xlim([min(graph.x_of_node) - 0.5, max(graph.x_of_node) + 0.5])
    plt.ylim([min(graph.y_of_node) - 0.5, max(graph.y_of_node) + 0.5])

    if "node" in locs:
        plot_nodes(graph, with_id=with_id, markersize=10)
    if "link" in locs:
        plot_links(graph, with_id=with_id, linewidth=None, as_arrow=False)
    if "patch" in locs:
        plot_patches(graph)

    if "corner" in locs:
        plot_nodes(graph.dual, color="c")
    if "face" in locs:
        plot_links(graph.dual, linestyle="dotted", color="k")
    if "cell" in locs:
        plot_patches(graph.dual, color="m")

    plt.xlabel("x")
    plt.ylabel("y")
    plt.gca().set_aspect(1.0)

    plt.show()



import numpy as np
from six.moves import range

from ..core.utils import make_optional_arg_into_id_array


def _swap(a, b):
    return (b, a)


def _iround(x):
    return int(round(x))


def neighbor_active_link_at_cell(grid, inds, *args):
    """neighbor_active_link_at_cell(grid, link_ids [, cell_ids])

    Return an array of the active link ids for neighbors of *cell_id* cells.
    *link_ids* is an index into the links of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    link_inds : array_like
        IDs of links
    cell_ids : array_like, optional
        IDs of cells for which to get links

    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]
    links = grid._active_links_at_node(node_ids).T

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return links[range(len(cell_ids)), inds]


def neighbor_node_at_cell(grid, inds, *args):
    """ node_id_of_cell_neighbor(grid, neighbor_ids [, cell_ids])

    Return an array of the node ids for neighbors of *cell_id* cells.
    *neighbor_ids* is an index into the neighbors of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    neighbor_ids : array_like
        IDs of the neighbor nodes.
    cell_ids : array_like, optional
        IDs of cell about which to get neighbors.

    Returns
    -------
    ndarray
        Node IDs for given neighbors of cells.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_funcs import neighbor_node_at_cell
    >>> grid = RasterModelGrid((4, 5), xy_spacing=1.0)
    >>> neighbor_node_at_cell(grid, 0, 0)
    array([1])

    Get the lower and the the upper neighbors for all the cells.

    >>> neighbor_node_at_cell(grid, 0)
    array([1, 2, 3, 6, 7, 8])
    >>> neighbor_node_at_cell(grid, 2)
    array([11, 12, 13, 16, 17, 18])

    As an alternative to the above, use fancy-indexing to get both sets of
    neighbors with one call.

    >>> neighbor_node_at_cell(grid, np.array([0, 2]), [1, 4])
    array([[ 2, 12],
           [ 7, 17]])
    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]
    neighbors = grid.active_adjacent_nodes_at_node[node_ids]

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    # return neighbors[range(len(cell_ids)), 3 - inds]
    return np.take(np.take(neighbors, range(len(cell_ids)), axis=0), 3 - inds, axis=1)


def calculate_slope_aspect_bfp(xs, ys, zs):
    """Calculate slope and aspect.

    .. codeauthor:: Katy Barnhart <katherine.barnhart@colorado.edu>

    Fits a plane to the given N points with given *xs*, *ys*, and *zs* values
    using single value decomposition.

    Returns a tuple of (*slope*, *aspect*) based on the normal vector to the
    best fit plane.

    .. note::

        This function does not check if the points fall on a line, rather
        than a plane.
    """
    if not len(xs) == len(ys) == len(zs):
        raise ValueError("array must be the same length")

    # step 1: subtract the centroid from the points
    # step 2: create a 3XN matrix of the points for SVD
    # in python, the unit normal to the best fit plane is
    # given by the third column of the U matrix.
    mat = np.vstack((xs - np.mean(xs), ys - np.mean(ys), zs - np.mean(zs)))
    U, _, _ = np.linalg.svd(mat)
    normal = U[:, 2]

    # step 3: calculate the aspect
    asp = 90.0 - np.degrees(np.arctan2(normal[1], normal[0]))
    asp = asp % 360.0

    # step 4: calculate the slope
    slp = 90.0 - np.degrees(np.arcsin(normal[2]))

    return slp, asp


def find_nearest_node(rmg, coords, mode="raise"):
    """Find the node nearest a point.

    Find the index to the node nearest the given x, y coordinates.
    Coordinates are provided as numpy arrays in the *coords* tuple.
    *coords* is tuple of coordinates, one for each dimension.

    The *mode* keyword to indicate what to do if a point is outside of the
    grid. Valid choices are the same as that used with the numpy function
    `ravel_multi_index`.

    A point is considered to be outside of the grid if it is outside the
    perimeter of the grid by one half of the grid spacing.

    Parameters
    ----------
    rmg : RasterModelGrid
        The source grid.
    coords : tuple
        Coordinates of point as (x, y)
    mode : {'raise', 'wrap', 'clip'}, optional
        Action to take if a point is outside of the grid.

    Returns
    -------
    array_like :
        Indices of the nodes nearest the given coordinates.

    Examples
    --------
    Create a grid of 4 by 5 nodes with unit spacing.

    >>> import landlab
    >>> from landlab.grid.raster_funcs import find_nearest_node
    >>> rmg = landlab.RasterModelGrid((4, 5))

    The points can be either a tuple of scalars or of arrays.

    >>> find_nearest_node(rmg, (0.2, 0.6))
    5
    >>> find_nearest_node(rmg, (np.array([1.6, 3.6]), np.array([2.3, .7])))
    array([12,  9])

    The *mode* keyword indicates what to do if a point is outside of the
    grid.

    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='raise')
    Traceback (most recent call last):
        ...
    ValueError: invalid entry in coordinates array
    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='clip')
    5
    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='wrap')
    9
    """
    if isinstance(coords[0], np.ndarray):
        return _find_nearest_node_ndarray(rmg, coords, mode=mode)
    else:
        return find_nearest_node(
            rmg, (np.array(coords[0]), np.array(coords[1])), mode=mode
        )


def _find_nearest_node_ndarray(rmg, coords, mode="raise"):
    """Find the node nearest to a point.

    Parameters
    ----------
    rmg : RasterModelGrid
        A RasterModelGrid.
    coords : tuple of float
        Coordinates of test points as *x*, then *y*.
    mode : {'raise', 'wrap', 'clip'}, optional
        What to do with out-of-bounds indices (as with
        numpy.ravel_multi_index).

    Returns
    -------
    ndarray
        Nodes that are closest to the points.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _find_nearest_node_ndarray
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> grid = RasterModelGrid((4, 5))
    >>> _find_nearest_node_ndarray(grid, (.25, 1.25))
    5
    >>> _find_nearest_node_ndarray(grid, (.75, 2.25))
    11

    >>> grid = RasterModelGrid((4, 5), xy_spacing=(3, 4))
    >>> _find_nearest_node_ndarray(grid, (3.1, 4.1))
    6
    """
    column_indices = np.int_(np.around((coords[0] - rmg.node_x[0]) / rmg.dx))
    row_indices = np.int_(np.around((coords[1] - rmg.node_y[0]) / rmg.dy))

    return rmg.grid_coords_to_node_id(row_indices, column_indices, mode=mode)


def _value_is_in_bounds(value, bounds):
    """Check if a value is within bounds.

    Parameters
    ----------
    value : float or ndarray
        The test value.
    bounds : (lower, upper)
        The lower and upper bounds.

    Returns
    -------
    bool
        ``True`` if the value is within the bounds. Otherwise, ``False``.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _value_is_in_bounds
    >>> import numpy as np
    >>> _value_is_in_bounds(.5, (0, 1))
    True
    >>> _value_is_in_bounds(1, (0, 1))
    False
    >>> _value_is_in_bounds(0, (0, 1))
    True
    >>> _value_is_in_bounds(np.array((0, 1)), (0, 1))
    array([ True, False], dtype=bool)
    """
    dummy = value >= bounds[0]
    dummy &= value < bounds[1]
    return dummy


def _value_is_within_axis_bounds(rmg, value, axis):
    """Check if a value is within the bounds of a grid axis.

    Parameters
    ----------
    rmg : RasterModelGrid
        A RasterModelGrid.
    value : float
        The test value.
    axis : int
        The axis.

    Returns
    -------
    bool
        ``True`` if the value is within the axis bounds. Otherwise, ``False``.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _value_is_within_axis_bounds
    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid((4, 5))
    >>> _value_is_within_axis_bounds(rmg, 3.1, 0)
    False
    >>> _value_is_within_axis_bounds(rmg, 2.9, 0)
    True
    >>> _value_is_within_axis_bounds(rmg, 4.1, 1)
    False
    >>> _value_is_within_axis_bounds(rmg, 3.9, 1)
    True
    """
    axis_coord = rmg.node_axis_coordinates(axis)
    return _value_is_in_bounds(value, (axis_coord[0], axis_coord[-1]))


def is_coord_on_grid(rmg, coords, axes=(0, 1)):
    """Check if coordinates are contained on a grid.

    Parameters
    ----------
    rmg : RasterModelGrid
        Source grid.
    coords : tuple
        Coordinates of point as (x, y)
    axes : tuple, optional
        Check bounds only on a particular axis

    Examples
    --------
    Create a grid that ranges from x=0 to x=4, and y=0 to y=3.

    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_funcs import is_coord_on_grid
    >>> grid = RasterModelGrid((4, 5))
    >>> is_coord_on_grid(grid, (3.999, 2.999))
    True

    Check two points with one call. Numpy broadcasting rules apply for the
    point coordinates.

    >>> is_coord_on_grid(grid, ([3.9, 4.1], 2.9))
    array([ True, False], dtype=bool)

    >>> is_coord_on_grid(grid, ([3.9, 4.1], 2.9), axes=(0, ))
    array([ True,  True], dtype=bool)
    """
    coords = np.broadcast_arrays(*coords)

    is_in_bounds = _value_is_within_axis_bounds(rmg, coords[1 - axes[0]], axes[0])
    for axis in axes[1:]:
        is_in_bounds &= _value_is_within_axis_bounds(rmg, coords[1 - axis], axis)

    return is_in_bounds


def line_to_grid_coords(c0, r0, c1, r1):
    """Return integer grid coords forming line segment (c0, r0)->(c1, r1).

    Parameters
    ----------
    c0, r0 : int
        column and row coordinates of "starting" endpoint
    c1, r1 : int
        column and row coordinates of "ending" endpoint

    Returns
    -------
    rr, cc : (N,) ndarray of int
        row and column coordinates of nodes in the line

    Examples
    --------
    >>> line_to_grid_coords(0, 0, 4, 1)
    (array([0, 0, 0, 1, 1]), array([0, 1, 2, 3, 4]))

    Notes
    -----
    Inputs must be grid coordinates rather than actual (x, y) values (unless
    the grid has unit spacing, in which case they are the same). To convert
    from real (x, y) to (x_grid, y_grid) use x_grid = x / Dx, where Dx is
    horizontal grid spacing (and similarly for y).
        To convert a raster-grid node ID to column and row coords, use
    numpy.unravel_index(node_id, (num_rows, num_cols)).
        To convert the returned grid coordinates to node IDs, use the
    RasterModelGrid method grid_coords_to_node_id().
        This function uses an incremental algorithm for line scan-conversion
    (see, e.g., Foley et al., 1990, chapter 3). For a line with a slope
    0 > m > 1, start with the x coordinates for a set of grid columns that span
    the line. The corresponding y of the first one is just y0. The y for the
    next one is y0 + m, for the next y0 + 2m, etc. If m > 1, you use y instead
    of x. In the below, any line segments that "point" toward the lower-left
    half-grid (i.e., with azimuth between 135o and 315o) have their endpoints
    flipped first.
    """

    dx = c1 - c0
    dy = r1 - r0

    # Flip endpoints if needed to have segment point to up/right
    if (dx + dy) < 0:
        (c0, c1) = _swap(c0, c1)
        (r0, r1) = _swap(r0, r1)
        dx = -dx
        dy = -dy
        flip_array = True
    else:
        flip_array = False

    if dx > dy:  # more horizontal than vertical
        npts = _iround(c1 - c0) + 1
        cc = np.zeros(npts, dtype=int)
        rr = np.zeros(npts, dtype=int)
        cc[:] = np.arange(npts)
        rr[:] = np.round(r0 + (float(dy) / dx) * cc)
        cc[:] += _iround(c0)
    else:
        npts = _iround(r1 - r0) + 1
        cc = np.zeros(npts, dtype=int)
        rr = np.zeros(npts, dtype=int)
        rr[:] = np.arange(npts)
        cc[:] = np.round(c0 + (float(dx) / dy) * rr)
        rr[:] += _iround(r0)

    # If endpoints were flipped, here we "un-flip" again
    if flip_array:
        rr = np.flipud(rr)
        cc = np.flipud(cc)

    return rr, cc



import numpy as np
from pytest import approx

from landlab.utils import add_halo


def test_add_halo_with_defaults():
    values = np.arange(6, dtype=int).reshape((2, 3))
    with_halo = add_halo(values)
    assert np.all(with_halo[1:-1, 1:-1] == values)
    assert values.dtype == with_halo.dtype


def test_add_halo_to_float_array():
    values = np.arange(6, dtype=float).reshape((2, 3))
    with_halo = add_halo(values)
    assert np.all(with_halo[1:-1, 1:-1] == approx(values))
    assert values.dtype == with_halo.dtype


def test_add_halo_to_bool_array():
    values = np.full((2, 3), True, dtype=bool)
    with_halo = add_halo(values)
    assert np.all(with_halo[1:-1, 1:-1] == values)
    assert values.dtype == with_halo.dtype


def test_add_halo_with_zero_halo():
    values = np.ones((2, 3), dtype=int)
    with_halo = add_halo(values, halo=0)
    assert np.all(with_halo == values)


def test_add_halo_with_bigger_halo():
    values = np.ones((2, 4), dtype=int)
    with_halo = add_halo(values, halo=3, halo_value=0)
    assert np.all(
        with_halo
        == [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
            [0, 0, 0, 1, 1, 1, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )


def test_add_halo_where_halo_value_is_zero():
    values = np.ones((3, 4), dtype=int)
    with_halo = add_halo(values, halo_value=0)
    assert np.all(
        with_halo
        == [
            [0, 0, 0, 0, 0, 0],
            [0, 1, 1, 1, 1, 0],
            [0, 1, 1, 1, 1, 0],
            [0, 1, 1, 1, 1, 0],
            [0, 0, 0, 0, 0, 0],
        ]
    )


def test_add_halo_where_halo_value_is_nan():
    values = np.ones((2, 3), dtype=float)
    with_halo = add_halo(values, halo=1, halo_value=np.nan)
    assert np.all(with_halo[1:-1, 1:-1] == approx(values))
    assert np.all(
        np.isnan(with_halo)
        == [
            [True, True, True, True, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, True, True, True, True],
        ]
    )



#! /usr/env/python

"""
flow_director_dinf.py: provides the component FlowDirectorDINF.

Directs flow on raster grids only using the Dinfinity algorithm of
Tarboton 1997.
"""

import numpy

from landlab import (
    BAD_INDEX_VALUE,
    FIXED_GRADIENT_BOUNDARY,
    FIXED_VALUE_BOUNDARY,
    VoronoiDelaunayGrid,
)
from landlab.components.flow_director import flow_direction_dinf
from landlab.components.flow_director.flow_director_to_many import _FlowDirectorToMany


class FlowDirectorDINF(_FlowDirectorToMany):

    """Flow direction on a raster grid by the D infinity method.

    Directs flow by the D infinity method (Tarboton, 1997). Each node is
    assigned two flow directions, toward the two neighboring nodes that are on
    the steepest subtriangle. Partitioning of flow is done based on the aspect
    of the subtriangle.

    Specifically, it stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_node'*. This array is 2D, and is
       of dimension (number of nodes x max number of receivers).
    -  Node array of flow proportions: *'flow__receiver_proportions'*. This
       array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of links carrying flow:  *'flow__link_to_receiver_node'*.
       This array is 2D, and is of dimension (number of nodes x max number of
       receivers).
    -  Node array of downhill slopes corresponding to each receiver.:
       *'topographic__steepest_slope'* This array is 2D, and is
       of dimension (number of nodes x max number of receivers). If the slope is
       uphill or flat, the value is assigned zero.
    -  Boolean node array of all local lows: *'flow__sink_flag'*
    -  Link array identifing if flow goes with (1) or against (-1) the link
       direction: *'flow_link_direction'*

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------

    This method works for both raster and irregular grids. First we will look
    at a raster example, and then an irregular example.

    >>> import numpy as numpy
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorDINF
    >>> mg = RasterModelGrid((4,4), xy_spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x**2 + mg.node_y**2,
    ...     at = 'node'
    ... )

    The DINF flow director can be uses for raster grids only.

    >>> fd = FlowDirectorDINF(mg, 'topographic__elevation')
    >>> fd.surface_values.reshape(mg.shape)
    array([[  0.,   1.,   4.,   9.],
           [  1.,   2.,   5.,  10.],
           [  4.,   5.,   8.,  13.],
           [  9.,  10.,  13.,  18.]])
    >>> fd.run_one_step()

    Unlike flow directors that only direct flow to one node or to all
    downstream nodes, FlowDirectorDINF directs flow two nodes only. It stores
    the receiver information is a (number of nodes x 2) shape field at nodes.

    >>> mg.at_node['flow__receiver_node']
    array([[ 0, -1],
           [ 1, -1],
           [ 2, -1],
           [ 3, -1],
           [ 4, -1],
           [ 0,  -1],
           [ 5,  1],
           [ 7, -1],
           [ 8, -1],
           [ 5, -1],
           [ 5, -1],
           [11, -1],
           [12, -1],
           [13, -1],
           [14, -1],
           [15, -1]])

    It also stores the proportions of flow going to each receiver, the link on
    which the flow moves in at node arrays, and the slope of each link.

    >>> mg.at_node['flow__receiver_proportions'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.59033447,  0.40966553],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ]])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([[-1, -1],
           [-1, -1],
           [-1, -1],
           [-1, -1],
           [ 3, 25],
           [24,  4],
           [ 8, 26],
           [ 9, 28],
           [14, 31],
           [11, 33],
           [32, 12],
           [16, 34],
           [21, 37],
           [18, 39],
           [19, 38],
           [20, 40]])
    >>> mg.at_node['topographic__steepest_slope'] # doctest: +NORMALIZE_WHITESPACE
    array([[ -1.00000000e+00,  -1.41421356e+00],
           [  1.00000000e+00,  -7.12763635e+02],
           [  3.00000000e+00,   1.41421356e+00],
           [  5.00000000e+00,   2.82842712e+00],
           [  1.00900000e+03,   7.12763635e+02],
           [  1.41421356e+00,   1.00000000e+00],
           [  3.00000000e+00,   2.82842712e+00],
           [  1.00400000e+03,   7.10642315e+02],
           [  1.00400000e+03,   7.12056529e+02],
           [  3.00000000e+00,   0.00000000e+00],
           [  4.24264069e+00,   3.00000000e+00],
           [  1.00100000e+03,   7.09935208e+02],
           [ -0.00000000e+00,   7.09935208e+02],
           [  1.00400000e+03,   7.07813888e+02],
           [  1.00100000e+03,   7.09935208e+02],
           [  0.00000000e+00,   7.07813888e+02]])

    Finally, FlowDirectorDINF identifies sinks, or local lows.

    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1], dtype=int8)

    The flow directors also have the ability to return the flow receiver nodes
    through a function called direct_flow()

    >>> fd = FlowDirectorDINF(mg, 'topographic__elevation')
    >>> fd.run_one_step()
    >>> receivers, proportions = fd.direct_flow()
    >>> receivers
    array([[ 0, -1],
           [ 1, -1],
           [ 2, -1],
           [ 3, -1],
           [ 4, -1],
           [ 0, -1],
           [ 5,  1],
           [ 7, -1],
           [ 8, -1],
           [ 5, -1],
           [ 5, -1],
           [11, -1],
           [12, -1],
           [13, -1],
           [14, -1],
           [15, -1]])
    >>> proportions # doctest: +NORMALIZE_WHITESPACE
    array([[ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 0.59033447,  0.40966553],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ],
           [ 1.        ,  0.        ]])

    For each donor node (represented by each row) the proportions should sum to
    one.

    >>> proportions.sum(axis=1)
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,
            1.,  1.,  1.])
    """

    _name = "FlowDirectorDINF"

    def __init__(self, grid, surface="topographic__elevation"):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        surface : field name at node or array of length node, optional
            The surface to direct flow across, default is field at node:
            topographic__self.surface_valuesation.
        partition_method: string, optional
            Method for partitioning flow. Options include 'slope' (default) and
            'square_root_of_slope'.
        """

        self.method = "DINF"
        self.max_receivers = 2
        super(FlowDirectorDINF, self).__init__(grid, surface)
        self._is_Voroni = isinstance(self._grid, VoronoiDelaunayGrid)
        if self._is_Voroni:
            raise NotImplementedError(
                "FlowDirectorDINF is not implemented" " for irregular grids."
            )

        self.updated_boundary_conditions()

        # set the number of recievers, proportions, and receiver links with the
        # right size.
        self.receivers = grid.add_field(
            "flow__receiver_node",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=int),
            at="node",
            dtype=int,
            noclobber=False,
        )

        self.steepest_slope = grid.add_field(
            "topographic__steepest_slope",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=float),
            at="node",
            dtype=float,
            noclobber=False,
        )

        self.receiver_links = grid.add_field(
            "flow__link_to_receiver_node",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=int),
            at="node",
            dtype=int,
            noclobber=False,
        )

        self.proportions = grid.add_field(
            "flow__receiver_proportions",
            BAD_INDEX_VALUE
            * numpy.ones((self._grid.number_of_nodes, self.max_receivers), dtype=float),
            at="node",
            dtype=int,
            noclobber=False,
        )

    def updated_boundary_conditions(self):
        """Method to update FlowDirectorDINF when boundary conditions change.

        Call this if boundary conditions on the grid are updated after
        the component is instantiated.
        """
        self._active_links = self.grid.active_links
        self._activelink_tail = self.grid.node_at_link_tail[self.grid.active_links]
        self._activelink_head = self.grid.node_at_link_head[self.grid.active_links]

    def run_one_step(self):
        """Find flow directions and save to the model grid.

        run_one_step() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the status at node,
        calculates flow directions, and saves results to the grid.

        An alternative to direct_flow() is direct_flow() which does the same
        things but also returns the receiver nodes not return values.
        """
        self.direct_flow()

    def direct_flow(self):
        """Find flow directions, save to the model grid, and return receivers.

        direct_flow() checks for updated boundary conditions, calculates
        slopes on links, finds basself.surface_valuesel nodes based on the status at node,
        calculates flow directions, saves results to the grid, and returns a
        at-node array  of receiver nodes. This array is stored in the grid at:
        grid['node']['flow__receiver_nodes']

        An alternative to direct_flow() is run_one_step() which does the same
        things but also returns a at-node array  of receiver nodes. This array
        is stored in the grid at:
        grid['node']['flow__receiver_nodes']
        """
        self._check_updated_bc()

        # Step 1. Find and save base level nodes.
        (baselevel_nodes,) = numpy.where(
            numpy.logical_or(
                self._grid.status_at_node == FIXED_VALUE_BOUNDARY,
                self._grid.status_at_node == FIXED_GRADIENT_BOUNDARY,
            )
        )

        # Calculate flow directions
        (
            self.receivers,
            self.proportions,
            slopes_to_receivers,
            steepest_slope,
            steepest_receiver,
            sink,
            receiver_links,
            steepest_link,
        ) = flow_direction_dinf.flow_directions_dinf(
            self._grid, self.surface_values, baselevel_nodes=baselevel_nodes
        )

        # Save the four ouputs of this component.
        self._grid["node"]["flow__receiver_node"][:] = self.receivers
        self._grid["node"]["flow__receiver_proportions"][:] = self.proportions
        self._grid["node"]["topographic__steepest_slope"][:] = slopes_to_receivers
        self._grid["node"]["flow__link_to_receiver_node"][:] = receiver_links
        self._grid["node"]["flow__sink_flag"][:] = False
        self._grid["node"]["flow__sink_flag"][sink] = True

        return (self.receivers, self.proportions)


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 08:32:48 2016

@author: RCGlade
"""

import numpy as np

from landlab import Component


class ExponentialWeatherer(Component):

    r"""
    This component implements exponential weathering of bedrock on hillslopes.
    Uses exponential soil production function in the style of Ahnert (1976).

    Consider that :math:`w_0` is the maximum soil production decay depth and
    that :math:`w^*` is the characteristic soil production decay depth. The
    soil production rate :math:`\dot{w}` is given as a function of the soil
    depth :math:`d`,

    .. math::

        \dot{w} = w_0 \exp{\frac{d}{w_*}} \;.

    The `ExponentialWeatherer` only calculates soil production at core nodes.

    Parameters
    ----------
    grid: ModelGrid
        Landlab ModelGrid object
    soil_production__maximum_rate : float
        Characteristic weathering depth
    soil_production__decay_depth : float
        Maximum weathering rate for bare bedrock

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeatherer
    >>> mg = RasterModelGrid((5, 5))
    >>> soilz = mg.add_zeros('node', 'soil__depth')
    >>> soilrate = mg.add_ones('node', 'soil_production__rate')
    >>> expw = ExponentialWeatherer(mg)
    >>> expw.calc_soil_prod_rate()
    >>> np.allclose(mg.at_node['soil_production__rate'], 1.)
    True
    """

    _name = "ExponentialWeatherer"

    _input_var_names = ("soil__depth",)

    _output_var_names = ("soil_production__rate",)

    _var_units = {"soil__depth": "m", "soil_production__rate": "m/yr"}

    _var_mapping = {"soil__depth": "node", "soil_production__rate": "node"}

    _var_doc = {
        "soil__depth": "depth of soil/weather bedrock",
        "soil_production__rate": "rate of soil production at nodes",
    }

    def __init__(
        self,
        grid,
        soil_production__maximum_rate=1.0,
        soil_production__decay_depth=1.0,
        **kwds
    ):

        # Store grid and parameters
        self._grid = grid
        self.wstar = soil_production__decay_depth
        self.w0 = soil_production__maximum_rate

        # Create fields:
        # soil depth
        if "soil__depth" in grid.at_node:
            self.depth = grid.at_node["soil__depth"]
        else:
            self.depth = grid.add_zeros("node", "soil__depth")

        # weathering rate
        if "soil_production__rate" in grid.at_node:
            self.soil_prod_rate = grid.at_node["soil_production__rate"]
        else:
            self.soil_prod_rate = grid.add_zeros("node", "soil_production__rate")

    def calc_soil_prod_rate(self, **kwds):
        """Calculate soil production rate.
        """

        # apply exponential function
        self.soil_prod_rate[self._grid.core_nodes] = self.w0 * np.exp(
            -self.depth[self._grid.core_nodes] / self.wstar
        )

        # weather
        # self.weather[self._active_nodes] = (self.wnot*np.exp(-self.depth[self._active_nodes]/self.wstar))

    def run_one_step(self, dt=None, **kwds):
        """

        Parameters
        ----------
        dt: float
            Used only for compatibility with standard run_one_step.
        """
        self.calc_soil_prod_rate(**kwds)



import numpy as np
from pytest import approx, raises

from landlab.components.flexure import get_flexure_parameter, subside_point_load


def test_calc_flexure_parameter():
    eet, youngs = 65000.0, 7e10
    alpha_1d = get_flexure_parameter(eet, youngs, 1)
    alpha_2d = get_flexure_parameter(eet, youngs, 2)
    assert alpha_2d == approx(84828.72)
    assert alpha_1d == approx(np.power(4.0, 0.25) * alpha_2d)


def test_calc_flexure_parameter_bad_ndim():
    with raises(ValueError):
        get_flexure_parameter(65000.0, 7e10, 0)
    with raises(ValueError):
        get_flexure_parameter(65000.0, 7e10, 3)


def test_subside_point_load():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9
    loc = (5000.0, 2500.0)

    x = np.arange(0, 10000, 1000.0)
    y = np.arange(0, 5000, 1000.0)
    (x, y) = np.meshgrid(x, y)
    x.shape = (x.size,)
    y.shape = (y.size,)

    dz_one_load = subside_point_load(load, loc, (x, y), params=params)

    n_loads = 16

    dz = subside_point_load(
        np.full(n_loads, load / n_loads),
        np.full((n_loads, 2), loc).T,
        (x, y),
        params=params,
    )

    assert dz.mean() == approx(dz_one_load.mean())
    assert dz.min() == approx(dz_one_load.min())
    assert dz.max() == approx(dz_one_load.max())


def test_point_load_1d_with_scalar_args():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9

    x = np.arange(0, 10000, 1000.0)

    dz = subside_point_load(load, 5000.0, x, params=params)

    assert dz.shape == x.shape


def test_point_load_1d_is_symetric():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9

    x = np.arange(0, 11000, 1000.0)
    i_mid = (len(x) - 1) // 2

    dz = subside_point_load(load, x[i_mid], x, params=params)

    assert np.argmax(dz) == i_mid
    assert dz[i_mid] > 0.0
    assert np.all(dz[i_mid::-1] == dz[i_mid:])


def test_point_load_2d_is_symetric():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9

    n = 11
    n_mid = (n - 1) // 2

    (x, y) = np.meshgrid(np.arange(0.0, n), np.arange(0.0, n))

    dz = subside_point_load(
        load,
        (x[n_mid, n_mid], y[n_mid, n_mid]),
        (x.flatten(), y.flatten()),
        params=params,
    )
    dz.shape = (n, n)

    assert np.argmax(dz) == np.ravel_multi_index((n_mid, n_mid), (n, n))
    assert dz[n_mid, n_mid] > 0.0
    assert np.all(dz[:, n_mid::-1] == dz[:, n_mid:])
    assert np.all(dz[n_mid::-1, :] == dz[n_mid:, :])


def test_subside_point_load_1d():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9
    loc = (5000.0,)

    x = np.arange(0, 10000, 1000.0)

    dz_one_load = subside_point_load(load, loc, (x,), params=params)

    n_loads = 32
    dz = subside_point_load(
        np.full(n_loads, load / n_loads), (np.full(n_loads, loc),), (x,), params=params
    )

    assert np.all(dz == approx(dz_one_load))


def test_out_keyword():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9
    loc = (5000.0,)

    x = np.arange(0, 10000, 1000.0)
    out = np.empty_like(x)

    dz = subside_point_load(load, loc, (x,), params=params, out=out)

    assert dz is out


def test_dimension_mismatch():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9
    loc = (5000.0,)

    x = np.arange(0, 10000, 1000.0)
    y = np.zeros_like(x)
    with raises(ValueError):
        subside_point_load(load, loc, (x, y), params=params)


def test_ndim_too_big():
    params = dict(eet=65000.0, youngs=7e10)
    load = 1e9
    loc = ((5000.0, 1000.0, 500.0),)

    x = np.arange(0, 10000, 1000.0)
    y = np.zeros_like(x)
    z = np.zeros_like(x)
    with raises(ValueError):
        subside_point_load(load, loc, (x, y, z), params=params)



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_unit_spacing():
    """Test on a grid with unit spacing."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    grads = grid.calc_grad_at_link(values_at_nodes)
    assert_array_equal(
        grads,
        np.array(
            [
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
            ],
            dtype=float,
        ),
    )
    with pytest.deprecated_call():
        diffs = grid.calculate_diff_at_links(values_at_nodes)
    assert_array_equal(grads, diffs)


def test_non_unit_spacing():
    """Test on a grid with non-unit spacing."""
    grid = RasterModelGrid((4, 5), xy_spacing=(2, 5))
    values_at_nodes = np.arange(20.0)
    grads = grid.calc_grad_at_link(values_at_nodes)
    assert_array_equal(
        grads,
        np.array(
            [
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
            ],
            dtype=float,
        ),
    )
    diffs = grid.calc_diff_at_link(values_at_nodes)
    assert_array_equal(
        diffs,
        np.array(
            [
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
            ],
            dtype=float,
        ),
    )


def test_out_array():
    """Test using the out keyword."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    grads = np.empty(31)
    rtn_grads = grid.calc_grad_at_link(values_at_nodes, out=grads)
    assert_array_equal(
        grads,
        np.array(
            [
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
            ],
            dtype=float,
        ),
    )
    assert rtn_grads is grads


def test_diff_out_array():
    """Test that return array is the out keyword."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.0)
    diff = np.empty(31)
    with pytest.deprecated_call():
        rtn_diff = grid.calculate_diff_at_links(values_at_nodes, out=diff)
    assert_array_equal(
        diff,
        np.array(
            [
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
                5,
                5,
                5,
                5,
                5,
                1,
                1,
                1,
                1,
            ],
            dtype=float,
        ),
    )
    assert rtn_diff is diff



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 26 13:59:55 2019

@author: gtucker
"""

from numpy.testing import assert_array_equal

from landlab.grid.raster_funcs import line_to_grid_coords


def test_line_to_grid_coords():
    """Compass directions refer to orientation tested."""
    rr, cc = line_to_grid_coords(2, 0, 5, 0)  # E
    assert_array_equal(cc, [2, 3, 4, 5])
    assert_array_equal(rr, [0, 0, 0, 0])

    rr, cc = line_to_grid_coords(0, 0, 4, 1)  # ENE
    assert_array_equal(cc, [0, 1, 2, 3, 4])
    assert_array_equal(rr, [0, 0, 0, 1, 1])

    rr, cc = line_to_grid_coords(0, 2, 1, 5)  # NNE
    assert_array_equal(cc, [0, 0, 1, 1])
    assert_array_equal(rr, [2, 3, 4, 5])

    rr, cc = line_to_grid_coords(-2, 0, -2, 8)  # N
    assert_array_equal(cc, [-2, -2, -2, -2, -2, -2, -2, -2, -2])
    assert_array_equal(rr, [0, 1, 2, 3, 4, 5, 6, 7, 8])

    rr, cc = line_to_grid_coords(1, 1, 0, 5)  # NNW
    assert_array_equal(cc, [1, 1, 0, 0, 0])
    assert_array_equal(rr, [1, 2, 3, 4, 5])

    rr, cc = line_to_grid_coords(0, -1, -5, 0)  # WNW
    assert_array_equal(cc, [0, -1, -2, -3, -4, -5])
    assert_array_equal(rr, [-1, -1, -1, 0, 0, 0])

    rr, cc = line_to_grid_coords(0, -1, -6, -1)  # W
    assert_array_equal(cc, [0, -1, -2, -3, -4, -5, -6])
    assert_array_equal(rr, [-1, -1, -1, -1, -1, -1, -1])

    rr, cc = line_to_grid_coords(-7, -1, -13, -4)  # WSW
    assert_array_equal(cc, [-7, -8, -9, -10, -11, -12, -13])
    assert_array_equal(rr, [-1, -2, -2, -2, -3, -4, -4])

    rr, cc = line_to_grid_coords(9, -4, 6, -11)  # SSW
    assert_array_equal(cc, [9, 9, 8, 8, 7, 7, 6, 6])
    assert_array_equal(rr, [-4, -5, -6, -7, -8, -9, -10, -11])

    rr, cc = line_to_grid_coords(-2, 1, -2, -6)  # S
    assert_array_equal(cc, [-2, -2, -2, -2, -2, -2, -2, -2])
    assert_array_equal(rr, [1, 0, -1, -2, -3, -4, -5, -6])

    rr, cc = line_to_grid_coords(0, 0, 3, -6)  # SSE
    assert_array_equal(cc, [0, 0, 1, 2, 2, 2, 3])
    assert_array_equal(rr, [0, -1, -2, -3, -4, -5, -6])

    rr, cc = line_to_grid_coords(0, 0, 7, -3)  # ESE
    assert_array_equal(cc, [0, 1, 2, 3, 4, 5, 6, 7])
    assert_array_equal(rr, [0, 0, -1, -1, -2, -2, -3, -3])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 10:39:32 2017

@author: gtucker
"""
import pytest

from landlab import RasterModelGrid
from landlab.components import DepthDependentDiffuser, ExponentialWeatherer


def test_raise_kwargs_error():
    mg = RasterModelGrid((5, 5))
    soilTh = mg.add_zeros("node", "soil__depth")
    z = mg.add_zeros("node", "topographic__elevation")
    BRz = mg.add_zeros("node", "bedrock__elevation")
    z += mg.node_x.copy()
    BRz += mg.node_x / 2.0
    soilTh[:] = z - BRz
    ExponentialWeatherer(mg)
    with pytest.raises(TypeError):
        DepthDependentDiffuser(mg, diffusivity=1)

    DDdiff = DepthDependentDiffuser(mg)

    with pytest.raises(TypeError):
        DDdiff.soilflux(2.0, bad_var=1)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid

_GRIDS = {}


def setup_grids():
    """Set up test grids with unit and non-unit spacing."""
    _GRIDS.update(
        {
            "unit": RasterModelGrid((4, 5)),
            "non_unit": RasterModelGrid((4, 5), xy_spacing=2.0),
            "non_square": RasterModelGrid((4, 5), xy_spacing=(5, 2)),
        }
    )


def test_unit_spacing():
    """Test with a grid with unit spacing."""
    rmg = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20)
    grads = rmg.calc_grad_at_link(values_at_nodes)[rmg.active_links]

    assert_array_equal(
        grads,
        np.array(
            [
                5.0,
                5.0,
                5.0,
                1.0,
                1.0,
                1.0,
                1.0,
                5.0,
                5.0,
                5.0,
                1.0,
                1.0,
                1.0,
                1.0,
                5.0,
                5.0,
                5.0,
            ]
        ),
    )

    diffs = rmg.calc_diff_at_link(values_at_nodes)[rmg.active_links]
    assert_array_equal(grads, diffs)


def test_non_unit_spacing():
    """Test with a grid with non-unit spacing."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2, 5))
    values_at_nodes = np.arange(20)

    grads = rmg.calc_grad_at_link(values_at_nodes)[rmg.active_links]
    assert_array_equal(
        grads,
        np.array(
            [
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
            ]
        ),
    )
    diffs = rmg.calc_diff_at_link(values_at_nodes)[rmg.active_links]
    assert_array_equal(
        diffs,
        np.array(
            [
                5.0,
                5.0,
                5.0,
                1.0,
                1.0,
                1.0,
                1.0,
                5.0,
                5.0,
                5.0,
                1.0,
                1.0,
                1.0,
                1.0,
                5.0,
                5.0,
                5.0,
            ]
        ),
    )


def test_out_array():
    """Test using the out keyword."""
    rmg = RasterModelGrid((4, 5), xy_spacing=(2, 5))
    values_at_nodes = np.arange(20)

    output_array = np.empty(rmg.number_of_links)
    rtn_array = rmg.calc_grad_at_link(values_at_nodes, out=output_array)
    assert_array_equal(
        rtn_array[rmg.active_links],
        np.array(
            [
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
                0.5,
                0.5,
                0.5,
                0.5,
                1.0,
                1.0,
                1.0,
            ]
        ),
    )
    assert rtn_array is output_array


def test_diff_out_array():
    """Test returned array is the same as that passed as out keyword."""
    rmg = RasterModelGrid((4, 5))
    values = np.arange(20)
    diff = np.empty(rmg.number_of_links)
    rtn_diff = rmg.calc_diff_at_link(values, out=diff)
    assert_array_equal(
        diff[rmg.active_links],
        np.array([5, 5, 5, 1, 1, 1, 1, 5, 5, 5, 1, 1, 1, 1, 5, 5, 5]),
    )
    assert rtn_diff is diff
